from __future__ import annotations

import copy
import logging
import random
from typing import Optional

import igraph as ig
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pydantic import BaseModel

from napistu import sbml_dfs_core
from napistu import utils
from napistu.network import net_create_utils
from napistu.network.ng_core import NapistuGraph


from napistu.constants import (
    MINI_SBO_FROM_NAME,
    SBO_MODIFIER_NAMES,
    SBOTERM_NAMES,
    SBML_DFS,
    ENTITIES_W_DATA,
)

from napistu.network.constants import (
    NAPISTU_GRAPH_VERTICES,
    NAPISTU_GRAPH_EDGES,
    NAPISTU_GRAPH_EDGE_DIRECTIONS,
    NAPISTU_GRAPH_NODE_TYPES,
    GRAPH_WIRING_APPROACHES,
    NAPISTU_WEIGHTING_STRATEGIES,
    VALID_GRAPH_WIRING_APPROACHES,
    VALID_WEIGHTING_STRATEGIES,
    DEFAULT_WT_TRANS,
    DEFINED_WEIGHT_TRANSFORMATION,
    SCORE_CALIBRATION_POINTS_DICT,
    SOURCE_VARS_DICT,
    DROP_REACTIONS_WHEN,
)


logger = logging.getLogger(__name__)


def create_napistu_graph(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    reaction_graph_attrs: Optional[dict] = None,
    directed: bool = True,
    edge_reversed: bool = False,
    wiring_approach: str = GRAPH_WIRING_APPROACHES.REGULATORY,
    drop_reactions_when: str = DROP_REACTIONS_WHEN.SAME_TIER,
    verbose: bool = False,
    custom_transformations: Optional[dict] = None,
) -> NapistuGraph:
    """
    Create a NapistuGraph network from a mechanistic network using one of a set of wiring approaches.

    Parameters
    ----------
    sbml_dfs : sbml_dfs_core.SBML_dfs
        A model formed by aggregating pathways.
    reaction_graph_attrs : dict, optional
        Dictionary containing attributes to pull out of reaction_data and a weighting scheme for the graph.
    directed : bool, optional
        Whether to create a directed (True) or undirected (False) graph. Default is True.
    edge_reversed : bool, optional
        Whether to reverse the directions of edges. Default is False.
    wiring_approach : str, optional
        Type of graph to create. Valid values are:
            - 'bipartite': substrates and modifiers point to the reaction they drive, this reaction points to products
            - 'regulatory': non-enzymatic modifiers point to enzymes, enzymes point to substrates and products
            - 'surrogate': non-enzymatic modifiers -> substrates -> enzymes -> reaction -> products
            - 'bipartite_og': old method for generating a true bipartite graph. Retained primarily for regression testing.
    drop_reactions_when : str, optional
        The condition under which to drop reactions as a network vertex. Valid values are:
            - 'same_tier': drop reactions when all participants are on the same tier of a wiring hierarchy
            - 'edgelist': drop reactions when the reaction species are only 2 (1 reactant + 1 product)
            - 'always': drop reactions regardless of tiers
    verbose : bool, optional
        Extra reporting. Default is False.
    custom_transformations : dict, optional
        Dictionary of custom transformation functions to use for attribute transformation.

    Returns
    -------
    NapistuGraph
        A NapistuGraph network (subclass of igraph.Graph).

    Raises
    ------
    ValueError
        If wiring_approach is not valid or if required attributes are missing.
    """

    if reaction_graph_attrs is None:
        reaction_graph_attrs = {}

    if wiring_approach not in VALID_GRAPH_WIRING_APPROACHES + ["bipartite_og"]:
        raise ValueError(
            f"wiring_approach is not a valid value ({wiring_approach}), valid values are {','.join(VALID_GRAPH_WIRING_APPROACHES)}"
        )

    # fail fast if reaction_graph_attrs is not properly formatted
    for k in reaction_graph_attrs.keys():
        _validate_entity_attrs(
            reaction_graph_attrs[k], custom_transformations=custom_transformations
        )

    working_sbml_dfs = copy.deepcopy(sbml_dfs)
    reaction_species_counts = working_sbml_dfs.reaction_species.value_counts(
        SBML_DFS.R_ID
    )
    valid_reactions = reaction_species_counts[reaction_species_counts > 1].index
    # due to autoregulation reactions, and removal of cofactors some
    # reactions may have 1 (or even zero) species. drop these.

    n_dropped_reactions = working_sbml_dfs.reactions.shape[0] - len(valid_reactions)
    if n_dropped_reactions != 0:
        logger.info(
            f"Dropping {n_dropped_reactions} reactions with <= 1 reaction species "
            "these underspecified reactions may be due to either unrepresented "
            "autoregulation and/or removal of cofactors."
        )

        working_sbml_dfs.reactions = working_sbml_dfs.reactions[
            working_sbml_dfs.reactions.index.isin(valid_reactions)
        ]
        working_sbml_dfs.reaction_species = working_sbml_dfs.reaction_species[
            working_sbml_dfs.reaction_species[SBML_DFS.R_ID].isin(valid_reactions)
        ]

    logger.info(
        "Organizing all network nodes (compartmentalized species and reactions)"
    )

    network_nodes = list()
    network_nodes.append(
        working_sbml_dfs.compartmentalized_species.reset_index()[
            [SBML_DFS.SC_ID, SBML_DFS.SC_NAME]
        ]
        .rename(columns={SBML_DFS.SC_ID: "node_id", SBML_DFS.SC_NAME: "node_name"})
        .assign(node_type=NAPISTU_GRAPH_NODE_TYPES.SPECIES)
    )
    network_nodes.append(
        working_sbml_dfs.reactions.reset_index()[[SBML_DFS.R_ID, SBML_DFS.R_NAME]]
        .rename(columns={SBML_DFS.R_ID: "node_id", SBML_DFS.R_NAME: "node_name"})
        .assign(node_type=NAPISTU_GRAPH_NODE_TYPES.REACTION)
    )

    # rename nodes to name since it is treated specially
    network_nodes_df = pd.concat(network_nodes).rename(
        columns={"node_id": NAPISTU_GRAPH_VERTICES.NAME}
    )

    logger.info(f"Formatting edges as a {wiring_approach} graph")

    if wiring_approach == "bipartite_og":
        network_edges = _create_napistu_graph_bipartite(working_sbml_dfs)
    elif wiring_approach in VALID_GRAPH_WIRING_APPROACHES:
        # pass wiring_approach so that an appropriate tiered schema can be used.
        network_edges = create_napistu_graph_wiring(
            working_sbml_dfs, wiring_approach, drop_reactions_when
        )
    else:
        raise NotImplementedError("Invalid wiring_approach")

    logger.info("Adding reversibility and other meta-data from reactions_data")
    augmented_network_edges = _augment_network_edges(
        network_edges,
        working_sbml_dfs,
        reaction_graph_attrs,
        custom_transformations=custom_transformations,
    )

    logger.info(
        "Creating reverse reactions for reversible reactions on a directed graph"
    )
    if directed:
        directed_network_edges = pd.concat(
            [
                # assign forward edges
                augmented_network_edges.assign(
                    **{
                        NAPISTU_GRAPH_EDGES.DIRECTION: NAPISTU_GRAPH_EDGE_DIRECTIONS.FORWARD
                    }
                ),
                # create reverse edges for reversible reactions
                _reverse_network_edges(augmented_network_edges),
            ]
        )
    else:
        directed_network_edges = augmented_network_edges.assign(
            **{NAPISTU_GRAPH_EDGES.DIRECTION: NAPISTU_GRAPH_EDGE_DIRECTIONS.UNDIRECTED}
        )

    # de-duplicate edges
    unique_edges = (
        directed_network_edges.groupby(
            [NAPISTU_GRAPH_EDGES.FROM, NAPISTU_GRAPH_EDGES.TO]
        )
        .first()
        .reset_index()
    )

    if unique_edges.shape[0] != directed_network_edges.shape[0]:
        logger.warning(
            f"{directed_network_edges.shape[0] - unique_edges.shape[0]} edges were dropped "
            "due to duplicated origin -> target relationiships, use verbose for "
            "more information"
        )

        if verbose:
            # report duplicated edges
            grouped_edges = directed_network_edges.groupby(
                [NAPISTU_GRAPH_EDGES.FROM, NAPISTU_GRAPH_EDGES.TO]
            )
            duplicated_edges = [
                grouped_edges.get_group(x)
                for x in grouped_edges.groups
                if grouped_edges.get_group(x).shape[0] > 1
            ]
            example_duplicates = pd.concat(
                random.sample(duplicated_edges, min(5, len(duplicated_edges)))
            )

            logger.warning(utils.style_df(example_duplicates, headers="keys"))

    # convert nodes and edgelist into an igraph network
    logger.info("Formatting cpr_graph output")
    napistu_ig_graph = ig.Graph.DictList(
        vertices=network_nodes_df.to_dict("records"),
        edges=unique_edges.to_dict("records"),
        directed=directed,
        vertex_name_attr=NAPISTU_GRAPH_VERTICES.NAME,
        edge_foreign_keys=(NAPISTU_GRAPH_EDGES.FROM, NAPISTU_GRAPH_EDGES.TO),
    )

    # delete singleton nodes (most of these will be reaction nodes associated with pairwise interactions)

    # Always return NapistuGraph
    napistu_graph = NapistuGraph.from_igraph(
        napistu_ig_graph, wiring_approach=wiring_approach, is_reversed=edge_reversed
    )

    # remove singleton nodes (mostly reactions that are not part of any interaction)
    napistu_graph.remove_isolated_vertices()

    if edge_reversed:
        logger.info("Applying edge reversal using reversal utilities")
        napistu_graph.reverse_edges()

    return napistu_graph


def process_napistu_graph(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    reaction_graph_attrs: Optional[dict] = None,
    directed: bool = True,
    edge_reversed: bool = False,
    wiring_approach: str = GRAPH_WIRING_APPROACHES.BIPARTITE,
    weighting_strategy: str = NAPISTU_WEIGHTING_STRATEGIES.UNWEIGHTED,
    verbose: bool = False,
    custom_transformations: dict = None,
) -> NapistuGraph:
    """
    Process Consensus Graph.

    Sets up a NapistuGraph network and then adds weights and other malleable attributes.

    Parameters
    ----------
    sbml_dfs : sbml_dfs_core.SBML_dfs
        A model formed by aggregating pathways.
    reaction_graph_attrs : dict, optional
        Dictionary containing attributes to pull out of reaction_data and a weighting scheme for the graph.
    directed : bool, optional
        Whether to create a directed (True) or undirected (False) graph. Default is True.
    edge_reversed : bool, optional
        Whether to reverse the directions of edges. Default is False.
    wiring_approach : str, optional
        Type of graph to create. See `create_napistu_graph` for valid values.
    weighting_strategy : str, optional
        A network weighting strategy. Options:
            - 'unweighted': all weights (and upstream_weight for directed graphs) are set to 1.
            - 'topology': weight edges by the degree of the source nodes favoring nodes with few connections.
            - 'mixed': transform edges with a quantitative score based on reaction_attrs; and set edges without quantitative score as a source-specific weight.
            - 'calibrated': transform edges with a quantitative score based on reaction_attrs and combine them with topology scores to generate a consensus.
    verbose : bool, optional
        Extra reporting. Default is False.
    custom_transformations : dict, optional
        Dictionary of custom transformation functions to use for attribute transformation.

    Returns
    -------
    NapistuGraph
        A weighted NapistuGraph network (subclass of igraph.Graph).
    """

    if reaction_graph_attrs is None:
        reaction_graph_attrs = {}

    logging.info("Constructing network")
    napistu_graph = create_napistu_graph(
        sbml_dfs,
        reaction_graph_attrs,
        directed=directed,
        edge_reversed=edge_reversed,
        wiring_approach=wiring_approach,
        verbose=verbose,
        custom_transformations=custom_transformations,
    )

    if "reactions" in reaction_graph_attrs.keys():
        reaction_attrs = reaction_graph_attrs["reactions"]
    else:
        reaction_attrs = dict()

    logging.info(f"Adding edge weights with an {weighting_strategy} strategy")

    weighted_napistu_graph = add_graph_weights(
        napistu_graph=napistu_graph,
        reaction_attrs=reaction_attrs,
        weighting_strategy=weighting_strategy,
    )

    return weighted_napistu_graph


def create_napistu_graph_wiring(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    wiring_approach: str,
    drop_reactions_when: str = DROP_REACTIONS_WHEN.SAME_TIER,
) -> pd.DataFrame:
    """
    Turn an sbml_dfs model into a tiered graph which links upstream entities to downstream ones.

    Parameters
    ----------
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The SBML_dfs object containing the model data.
    wiring_approach : str
        The wiring approach to use for the graph.
    drop_reactions_when : str, optional
        The condition under which to drop reactions as a network vertex. Default is 'same_tier'.

    Returns
    -------
    pd.DataFrame
        DataFrame representing the tiered network edges.

    Raises
    ------
    ValueError
        If invalid SBO terms are present or required attributes are missing.
    """

    # organize reaction species for defining connections
    logger.info(
        f"Turning {sbml_dfs.reaction_species.shape[0]} reactions species into edges."
    )

    all_reaction_edges_df = net_create_utils.wire_reaction_species(
        sbml_dfs.reaction_species, wiring_approach, drop_reactions_when
    )

    logger.info(
        "Adding additional attributes to edges, e.g., # of children and parents."
    )

    # add compartmentalized species summaries to weight edges
    cspecies_features = sbml_dfs.get_cspecies_features()

    # calculate undirected and directed degrees (i.e., # of parents and children)
    # based on a network's edgelist. this used when the network representation is
    # not the bipartite network which can be trivially obtained from the pathway
    # specification
    unique_edges = (
        all_reaction_edges_df.groupby(
            [NAPISTU_GRAPH_EDGES.FROM, NAPISTU_GRAPH_EDGES.TO]
        )
        .first()
        .reset_index()
    )

    # children
    n_children = (
        unique_edges[NAPISTU_GRAPH_EDGES.FROM]
        .value_counts()
        # rename values to the child name
        .to_frame(name=NAPISTU_GRAPH_EDGES.SC_CHILDREN)
        .reset_index()
        .rename(
            {
                NAPISTU_GRAPH_EDGES.FROM: SBML_DFS.SC_ID,
            },
            axis=1,
        )
    )

    # parents
    n_parents = (
        unique_edges[NAPISTU_GRAPH_EDGES.TO]
        .value_counts()
        # rename values to the parent name
        .to_frame(name=NAPISTU_GRAPH_EDGES.SC_PARENTS)
        .reset_index()
        .rename(
            {
                NAPISTU_GRAPH_EDGES.TO: SBML_DFS.SC_ID,
            },
            axis=1,
        )
    )

    graph_degree_by_edgelist = n_children.merge(n_parents, how="outer").fillna(int(0))

    graph_degree_by_edgelist[NAPISTU_GRAPH_EDGES.SC_DEGREE] = (
        graph_degree_by_edgelist[NAPISTU_GRAPH_EDGES.SC_CHILDREN]
        + graph_degree_by_edgelist[NAPISTU_GRAPH_EDGES.SC_PARENTS]
    )
    graph_degree_by_edgelist = (
        graph_degree_by_edgelist[
            ~graph_degree_by_edgelist[SBML_DFS.SC_ID].str.contains("R[0-9]{8}")
        ]
        .set_index(SBML_DFS.SC_ID)
        .sort_index()
    )

    cspecies_features = (
        cspecies_features.drop(
            [
                NAPISTU_GRAPH_EDGES.SC_DEGREE,
                NAPISTU_GRAPH_EDGES.SC_CHILDREN,
                NAPISTU_GRAPH_EDGES.SC_PARENTS,
            ],
            axis=1,
        )
        .join(graph_degree_by_edgelist)
        .fillna(int(0))
    )

    is_from_reaction = all_reaction_edges_df[NAPISTU_GRAPH_EDGES.FROM].isin(
        sbml_dfs.reactions.index.tolist()
    )
    is_from_reaction = all_reaction_edges_df[NAPISTU_GRAPH_EDGES.FROM].isin(
        sbml_dfs.reactions.index
    )
    # add substrate weight whenever "from" edge is a molecule
    # and product weight when the "from" edge is a reaction
    decorated_all_reaction_edges_df = pd.concat(
        [
            all_reaction_edges_df[~is_from_reaction].merge(
                cspecies_features, left_on=NAPISTU_GRAPH_EDGES.FROM, right_index=True
            ),
            all_reaction_edges_df[is_from_reaction].merge(
                cspecies_features, left_on=NAPISTU_GRAPH_EDGES.TO, right_index=True
            ),
        ]
    ).sort_index()

    if all_reaction_edges_df.shape[0] != decorated_all_reaction_edges_df.shape[0]:
        msg = (
            "'decorated_all_reaction_edges_df' and 'all_reaction_edges_df' should\n"
            "have the same number of rows but they did not"
        )

        raise ValueError(msg)

    logger.info(f"Done preparing {wiring_approach} graph")

    return decorated_all_reaction_edges_df


def pluck_entity_data(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    graph_attrs: dict[str, dict],
    data_type: str,
    custom_transformations: Optional[dict[str, callable]] = None,
) -> pd.DataFrame | None:
    """
    Pluck Entity Attributes from an sbml_dfs based on a set of tables and variables to look for.

    Parameters
    ----------
    sbml_dfs : sbml_dfs_core.SBML_dfs
        A mechanistic model.
    graph_attrs : dict
        A dictionary of species/reaction attributes to pull out. If the requested
        data_type ("species" or "reactions") is not present as a key, or if the value
        is an empty dict, this function will return None (no error).
    data_type : str
        "species" or "reactions" to pull out species_data or reactions_data.
    custom_transformations : dict[str, callable], optional
        A dictionary mapping transformation names to functions. If provided, these
        will be checked before built-in transformations. Example:
            custom_transformations = {"square": lambda x: x**2}

    Returns
    -------
    pd.DataFrame or None
        A table where all extracted attributes are merged based on a common index or None
        if no attributes were extracted. If the requested data_type is not present in
        graph_attrs, or if the attribute dict is empty, returns None. This is intended
        to allow optional annotation blocks.

    Raises
    ------
    ValueError
        If data_type is not valid or if requested tables/variables are missing.
    """

    if data_type not in ENTITIES_W_DATA:
        raise ValueError(
            f'"data_type" was {data_type} and must be in {", ".join(ENTITIES_W_DATA)}'
        )

    if data_type not in graph_attrs.keys():
        logger.info(
            f'No {data_type} annotations provided in "graph_attrs"; returning None'
        )
        return None

    entity_attrs = graph_attrs[data_type]
    # validating dict
    _validate_entity_attrs(entity_attrs, custom_transformations=custom_transformations)

    if len(entity_attrs) == 0:
        logger.info(
            f'No attributes defined for "{data_type}" in graph_attrs; returning None'
        )
        return None

    data_type_attr = data_type + "_data"
    entity_data_tbls = getattr(sbml_dfs, data_type_attr)

    data_list = list()
    for k, v in entity_attrs.items():
        # v["table"] is always present if entity_attrs is non-empty and validated
        if v["table"] not in entity_data_tbls.keys():
            raise ValueError(
                f"{v['table']} was defined as a table in \"graph_attrs\" but "
                f'it is not present in the "{data_type_attr}" of the sbml_dfs'
            )

        if v["variable"] not in entity_data_tbls[v["table"]].columns.tolist():
            raise ValueError(
                f"{v['variable']} was defined as a variable in \"graph_attrs\" but "
                f"it is not present in the {v['table']} of the \"{data_type_attr}\" of "
                "the sbml_dfs"
            )

        entity_series = entity_data_tbls[v["table"]][v["variable"]].rename(k)
        trans_name = v.get("trans", DEFAULT_WT_TRANS)
        # Look up transformation
        if custom_transformations and trans_name in custom_transformations:
            trans_fxn = custom_transformations[trans_name]
        elif trans_name in DEFINED_WEIGHT_TRANSFORMATION:
            trans_fxn = globals()[DEFINED_WEIGHT_TRANSFORMATION[trans_name]]
        else:
            # This should never be hit if _validate_entity_attrs is called correctly.
            raise ValueError(
                f"Transformation '{trans_name}' not found in custom_transformations or DEFINED_WEIGHT_TRANSFORMATION."
            )
        entity_series = entity_series.apply(trans_fxn)
        data_list.append(entity_series)

    if len(data_list) == 0:
        return None

    return pd.concat(data_list, axis=1)


def apply_weight_transformations(
    edges_df: pd.DataFrame, reaction_attrs: dict, custom_transformations: dict = None
):
    """
    Apply Weight Transformations to edge attributes.

    Parameters
    ----------
    edges_df : pd.DataFrame
        A table of edges and their attributes extracted from a cpr_graph.
    reaction_attrs : dict
        A dictionary of attributes identifying weighting attributes within
        an sbml_df's reaction_data, how they will be named in edges_df (the keys),
        and how they should be transformed (the "trans" aliases).
    custom_transformations : dict, optional
        A dictionary mapping transformation names to functions. If provided, these
        will be checked before built-in transformations.

    Returns
    -------
    pd.DataFrame
        edges_df with weight variables transformed.

    Raises
    ------
    ValueError
        If a weighting variable is missing or transformation is not found.
    """

    _validate_entity_attrs(
        reaction_attrs, custom_transformations=custom_transformations
    )

    transformed_edges_df = copy.deepcopy(edges_df)
    for k, v in reaction_attrs.items():
        if k not in transformed_edges_df.columns:
            raise ValueError(f"A weighting variable {k} was missing from edges_df")

        trans_name = v["trans"]
        # Look up transformation
        if custom_transformations and trans_name in custom_transformations:
            trans_fxn = custom_transformations[trans_name]
        elif trans_name in DEFINED_WEIGHT_TRANSFORMATION:
            trans_fxn = globals()[DEFINED_WEIGHT_TRANSFORMATION[trans_name]]
        else:
            # This should never be hit if _validate_entity_attrs is called correctly.
            raise ValueError(
                f"Transformation '{trans_name}' not found in custom_transformations or DEFINED_WEIGHT_TRANSFORMATION."
            )

        transformed_edges_df[k] = transformed_edges_df[k].apply(trans_fxn)

    return transformed_edges_df


def summarize_weight_calibration(
    napistu_graph: NapistuGraph, reaction_attrs: dict
) -> None:
    """
    Summarize Weight Calibration for a network with multiple sources for edge weights.

    Parameters
    ----------
    napistu_graph : NapistuGraph
        A graph where edge weights have already been calibrated.
    reaction_attrs : dict
        A dictionary summarizing the types of weights that exist and how they are transformed for calibration.

    Returns
    -------
    None
    """

    score_calibration_df = pd.DataFrame(SCORE_CALIBRATION_POINTS_DICT)
    score_calibration_df_calibrated = apply_weight_transformations(
        score_calibration_df, reaction_attrs
    )

    calibrated_edges = napistu_graph.get_edge_dataframe()

    _summarize_weight_calibration_table(
        calibrated_edges, score_calibration_df, score_calibration_df_calibrated
    )

    _summarize_weight_calibration_plots(
        calibrated_edges, score_calibration_df_calibrated
    )

    return None


def add_graph_weights(
    napistu_graph: NapistuGraph,
    reaction_attrs: dict,
    weighting_strategy: str = NAPISTU_WEIGHTING_STRATEGIES.UNWEIGHTED,
) -> NapistuGraph:
    """
    Add Graph Weights to a NapistuGraph using a specified weighting strategy.

    Parameters
    ----------
    napistu_graph : NapistuGraph
        A graphical network of molecules/reactions (nodes) and edges linking them (subclass of igraph.Graph).
    reaction_attrs : dict
        An optional dict of reaction attributes.
    weighting_strategy : str, optional
        A network weighting strategy. Options:
            - 'unweighted': all weights (and upstream_weight for directed graphs) are set to 1.
            - 'topology': weight edges by the degree of the source nodes favoring nodes emerging from nodes with few connections.
            - 'mixed': transform edges with a quantitative score based on reaction_attrs; and set edges without quantitative score as a source-specific weight.
            - 'calibrated': transform edges with a quantitative score based on reaction_attrs and combine them with topology scores to generate a consensus.

    Returns
    -------
    NapistuGraph
        The weighted NapistuGraph.

    Raises
    ------
    ValueError
        If weighting_strategy is not valid.
    """

    napistu_graph_updated = copy.deepcopy(napistu_graph)

    _validate_entity_attrs(reaction_attrs)

    if weighting_strategy not in VALID_WEIGHTING_STRATEGIES:
        raise ValueError(
            f"weighting_strategy was {weighting_strategy} and must be one of: "
            f"{', '.join(VALID_WEIGHTING_STRATEGIES)}"
        )

    # count parents and children and create weights based on them
    topology_weighted_graph = _create_topology_weights(napistu_graph_updated)

    if weighting_strategy == NAPISTU_WEIGHTING_STRATEGIES.TOPOLOGY:
        topology_weighted_graph.es[NAPISTU_GRAPH_EDGES.WEIGHT] = (
            topology_weighted_graph.es["topo_weights"]
        )
        if napistu_graph_updated.is_directed():
            topology_weighted_graph.es[NAPISTU_GRAPH_EDGES.UPSTREAM_WEIGHT] = (
                topology_weighted_graph.es["upstream_topo_weights"]
            )

        return topology_weighted_graph

    if weighting_strategy == NAPISTU_WEIGHTING_STRATEGIES.UNWEIGHTED:
        # set weights as a constant
        topology_weighted_graph.es[NAPISTU_GRAPH_EDGES.WEIGHT] = 1
        if napistu_graph_updated.is_directed():
            topology_weighted_graph.es[NAPISTU_GRAPH_EDGES.UPSTREAM_WEIGHT] = 1
        return topology_weighted_graph

    if weighting_strategy == NAPISTU_WEIGHTING_STRATEGIES.MIXED:
        return _add_graph_weights_mixed(topology_weighted_graph, reaction_attrs)

    if weighting_strategy == NAPISTU_WEIGHTING_STRATEGIES.CALIBRATED:
        return _add_graph_weights_calibration(topology_weighted_graph, reaction_attrs)

    raise ValueError(f"No logic implemented for {weighting_strategy}")


def _create_napistu_graph_bipartite(sbml_dfs: sbml_dfs_core.SBML_dfs) -> pd.DataFrame:
    """
    Turn an sbml_dfs model into a bipartite graph linking molecules to reactions.

    Parameters
    ----------
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The SBML_dfs object containing the model data.

    Returns
    -------
    pd.DataFrame
        DataFrame representing the bipartite network edges.
    """

    # setup edges
    network_edges = (
        sbml_dfs.reaction_species.reset_index()[
            [SBML_DFS.R_ID, SBML_DFS.SC_ID, SBML_DFS.STOICHIOMETRY, SBML_DFS.SBO_TERM]
        ]
        # rename species and reactions to reflect from -> to edges
        .rename(
            columns={
                SBML_DFS.SC_ID: NAPISTU_GRAPH_NODE_TYPES.SPECIES,
                SBML_DFS.R_ID: NAPISTU_GRAPH_NODE_TYPES.REACTION,
            }
        )
    )
    # add back an r_id variable so that each edge is annotated by a reaction
    network_edges[NAPISTU_GRAPH_EDGES.R_ID] = network_edges[
        NAPISTU_GRAPH_NODE_TYPES.REACTION
    ]

    # add edge weights
    cspecies_features = sbml_dfs.get_cspecies_features()
    network_edges = network_edges.merge(
        cspecies_features, left_on=NAPISTU_GRAPH_NODE_TYPES.SPECIES, right_index=True
    )

    # if directed then flip substrates and modifiers to the origin edge
    edge_vars = network_edges.columns.tolist()

    origins = network_edges[network_edges[SBML_DFS.STOICHIOMETRY] <= 0]
    origin_edges = origins.loc[:, [edge_vars[1], edge_vars[0]] + edge_vars[2:]].rename(
        columns={
            NAPISTU_GRAPH_NODE_TYPES.SPECIES: NAPISTU_GRAPH_EDGES.FROM,
            NAPISTU_GRAPH_NODE_TYPES.REACTION: NAPISTU_GRAPH_EDGES.TO,
        }
    )

    dests = network_edges[network_edges[SBML_DFS.STOICHIOMETRY] > 0]
    dest_edges = dests.rename(
        columns={
            NAPISTU_GRAPH_NODE_TYPES.REACTION: NAPISTU_GRAPH_EDGES.FROM,
            NAPISTU_GRAPH_NODE_TYPES.SPECIES: NAPISTU_GRAPH_EDGES.TO,
        }
    )

    network_edges = pd.concat([origin_edges, dest_edges])

    return network_edges


def _add_graph_weights_mixed(
    napistu_graph: NapistuGraph, reaction_attrs: dict
) -> NapistuGraph:
    """
    Weight a NapistuGraph using a mixed approach combining source-specific weights and existing edge weights.

    Parameters
    ----------
    napistu_graph : NapistuGraph
        The network to weight (subclass of igraph.Graph).
    reaction_attrs : dict
        Dictionary of reaction attributes to use for weighting.

    Returns
    -------
    NapistuGraph
        The weighted NapistuGraph.
    """

    edges_df = napistu_graph.get_edge_dataframe()

    calibrated_edges = apply_weight_transformations(edges_df, reaction_attrs)
    calibrated_edges = _create_source_weights(calibrated_edges, "source_wt")

    score_vars = list(reaction_attrs.keys())
    score_vars.append("source_wt")

    logger.info(f"Creating mixed scores based on {', '.join(score_vars)}")

    calibrated_edges[NAPISTU_GRAPH_EDGES.WEIGHT] = calibrated_edges[score_vars].min(
        axis=1
    )

    napistu_graph.es[NAPISTU_GRAPH_EDGES.WEIGHT] = calibrated_edges[
        NAPISTU_GRAPH_EDGES.WEIGHT
    ]
    if napistu_graph.is_directed():
        napistu_graph.es[NAPISTU_GRAPH_EDGES.UPSTREAM_WEIGHT] = calibrated_edges[
            NAPISTU_GRAPH_EDGES.WEIGHT
        ]

    # add other attributes and update transformed attributes
    napistu_graph.es["source_wt"] = calibrated_edges["source_wt"]
    for k in reaction_attrs.keys():
        napistu_graph.es[k] = calibrated_edges[k]

    return napistu_graph


def _add_graph_weights_calibration(
    napistu_graph: NapistuGraph, reaction_attrs: dict
) -> NapistuGraph:
    """
    Weight a NapistuGraph using a calibrated strategy which aims to roughly align qualitatively similar weights from different sources.

    Parameters
    ----------
    napistu_graph : NapistuGraph
        The network to weight (subclass of igraph.Graph).
    reaction_attrs : dict
        Dictionary of reaction attributes to use for weighting.

    Returns
    -------
    NapistuGraph
        The weighted NapistuGraph.
    """

    edges_df = napistu_graph.get_edge_dataframe()

    calibrated_edges = apply_weight_transformations(edges_df, reaction_attrs)

    score_vars = list(reaction_attrs.keys())
    score_vars.append("topo_weights")

    logger.info(f"Creating calibrated scores based on {', '.join(score_vars)}")
    napistu_graph.es[NAPISTU_GRAPH_EDGES.WEIGHT] = calibrated_edges[score_vars].min(
        axis=1
    )

    if napistu_graph.is_directed():
        score_vars = list(reaction_attrs.keys())
        score_vars.append("upstream_topo_weights")
        napistu_graph.es[NAPISTU_GRAPH_EDGES.UPSTREAM_WEIGHT] = calibrated_edges[
            score_vars
        ].min(axis=1)

    # add other attributes and update transformed attributes
    for k in reaction_attrs.keys():
        napistu_graph.es[k] = calibrated_edges[k]

    return napistu_graph


def _add_edge_attr_to_vertex_graph(
    napistu_graph: NapistuGraph,
    edge_attr_list: list,
    shared_node_key: str = "r_id",
) -> NapistuGraph:
    """
    Merge edge attribute(s) from edge_attr_list to vertices of a NapistuGraph.

    Parameters
    ----------
    napistu_graph : NapistuGraph
        A graph generated by create_napistu_graph() (subclass of igraph.Graph).
    edge_attr_list : list
        A list containing attributes to pull out of edges, then to add to vertices.
    shared_node_key : str, optional
        Key in edge that is shared with vertex, to map edge ids to corresponding vertex ids. Default is "r_id".

    Returns
    -------
    NapistuGraph
        The input NapistuGraph with additional vertex attributes added from edge attributes.
    """

    if len(edge_attr_list) == 0:
        logger.warning(
            "No edge attributes were passed, " "thus return the input graph."
        )
        return napistu_graph

    graph_vertex_df = napistu_graph.get_vertex_dataframe()
    graph_edge_df = napistu_graph.get_edge_dataframe()

    if shared_node_key not in graph_edge_df.columns.to_list():
        logger.warning(
            f"{shared_node_key} is not in the current edge attributes. "
            "shared_node_key must be an existing edge attribute"
        )
        return napistu_graph

    graph_edge_df_sub = graph_edge_df.loc[:, [shared_node_key] + edge_attr_list].copy()

    # check whether duplicated edge ids by shared_node_key have the same attribute values.
    # If not, give warning, and keep the first value. (which can be improved later)
    check_edgeid_attr_unique = (
        graph_edge_df_sub.groupby(shared_node_key)[edge_attr_list].nunique() == 1
    )

    # check any False in check_edgeid_attr_unique's columns, if so, get the column names
    bool_edgeid_attr_unique = (check_edgeid_attr_unique.isin([False])).any()  # type: ignore

    non_unique_indices = [
        i for i, value in enumerate(bool_edgeid_attr_unique.to_list()) if value
    ]

    # if edge ids with duplicated shared_node_key have more than 1 unique values
    # for attributes of interest
    non_unique_egde_attr = bool_edgeid_attr_unique.index[non_unique_indices].to_list()

    if len(non_unique_egde_attr) == 0:
        logger.info("Per duplicated edge ids, attributes have only 1 unique value.")
    else:
        logger.warning(
            f"Per duplicated edge ids, attributes: {non_unique_egde_attr} "
            "contain more than 1 unique values"
        )

    # remove duplicated edge attribute values
    graph_edge_df_sub_no_duplicate = graph_edge_df_sub.drop_duplicates(
        subset=shared_node_key, keep="first"
    )

    # rename shared_node_key to vertex key 'name'
    # as in net_create.create_napistu_graph(), vertex_name_attr is set to 'name'
    graph_edge_df_sub_no_duplicate = graph_edge_df_sub_no_duplicate.rename(
        columns={shared_node_key: "name"},
    )

    # merge edge attributes in graph_edge_df_sub_no_duplicate to vertex_df,
    # by shared key 'name'
    graph_vertex_df_w_edge_attr = pd.merge(
        graph_vertex_df,
        graph_edge_df_sub_no_duplicate,
        on="name",
        how="outer",
    )

    logger.info(f"Adding {edge_attr_list} to vertex attributes")
    # Warning for NaN values in vertex attributes:
    if graph_vertex_df_w_edge_attr.isnull().values.any():
        logger.warning(
            "NaN values are present in the newly added vertex attributes. "
            "Please assign proper values to those vertex attributes."
        )

    # assign the edge_attrs from edge_attr_list to napistu_graph's vertices:
    # keep the same edge attribute names:
    for col_name in edge_attr_list:
        napistu_graph.vs[col_name] = graph_vertex_df_w_edge_attr[col_name]

    return napistu_graph


def _summarize_weight_calibration_table(
    calibrated_edges: pd.DataFrame,
    score_calibration_df: pd.DataFrame,
    score_calibration_df_calibrated: pd.DataFrame,
):
    """
    Create a table comparing edge weights from multiple sources.

    Parameters
    ----------
    calibrated_edges : pd.DataFrame
        DataFrame of calibrated edge weights.
    score_calibration_df : pd.DataFrame
        DataFrame of raw calibration points.
    score_calibration_df_calibrated : pd.DataFrame
        DataFrame of calibrated calibration points.

    Returns
    -------
    pd.DataFrame
        Styled DataFrame summarizing calibration points and quantiles.
    """

    # generate a table summarizing different scoring measures
    #
    # a set of calibration points defined in DEFINED_WEIGHT_TRANSFORMATION which map
    # onto what we might consider strong versus dubious edges are compared to the
    # observed scores to see whether these calibration points generally map onto
    # the expected quantiles of the score distribution.
    #
    # different scores are also compared to see whether there calibrations are generally
    # aligned. that is to say a strong weight based on one scoring measure would receive
    # a similar quantitative score to a strong score for another measure.

    score_calibration_long_raw = (
        score_calibration_df.reset_index()
        .rename({"index": "edge_strength"}, axis=1)
        .melt(
            id_vars="edge_strength", var_name="weight_measure", value_name="raw_weight"
        )
    )

    score_calibration_long_calibrated = (
        score_calibration_df_calibrated.reset_index()
        .rename({"index": "edge_strength"}, axis=1)
        .melt(
            id_vars="edge_strength",
            var_name="weight_measure",
            value_name="trans_weight",
        )
    )

    score_calibration_table_long = score_calibration_long_raw.merge(
        score_calibration_long_calibrated
    )

    # compare calibration points to the quantiles of the observed score distributions
    score_quantiles = list()
    for ind, row in score_calibration_table_long.iterrows():
        score_quantiles.append(
            1
            - np.mean(
                calibrated_edges[row["weight_measure"]].dropna() >= row["trans_weight"]
            )
        )
    score_calibration_table_long["quantile_of_score_dist"] = score_quantiles

    return utils.style_df(score_calibration_table_long, headers="keys")


def _summarize_weight_calibration_plots(
    calibrated_edges: pd.DataFrame, score_calibration_df_calibrated: pd.DataFrame
) -> None:
    """
    Create plots summarizing the relationships between different scoring measures.

    Parameters
    ----------
    calibrated_edges : pd.DataFrame
        DataFrame of calibrated edge weights.
    score_calibration_df_calibrated : pd.DataFrame
        DataFrame of calibrated calibration points.

    Returns
    -------
    None
    """

    # set up a 2 x 1 plot
    f, (ax1, ax2) = plt.subplots(1, 2)

    calibrated_edges[["topo_weights", "string_wt"]].plot(
        kind="hist", bins=50, alpha=0.5, ax=ax1
    )
    ax1.set_title("Distribution of scores\npost calibration")

    score_calibration_df_calibrated.plot("weights", "string_wt", kind="scatter", ax=ax2)

    for k, v in score_calibration_df_calibrated.iterrows():
        ax2.annotate(k, v)
    ax2.axline((0, 0), slope=1.0, color="C0", label="by slope")
    ax2.set_title("Comparing STRING and\nTopology calibration points")

    return None


def _create_source_weights(
    edges_df: pd.DataFrame,
    source_wt_var: str = "source_wt",
    source_vars_dict: dict = SOURCE_VARS_DICT,
    source_wt_default: int = 1,
) -> pd.DataFrame:
    """
    Create weights based on an edge's source.

    Parameters
    ----------
    edges_df : pd.DataFrame
        The edges dataframe to add the source weights to.
    source_wt_var : str, optional
        The name of the column to store the source weights. Default is "source_wt".
    source_vars_dict : dict, optional
        Dictionary with keys indicating edge attributes and values indicating the weight to assign to that attribute. Default is SOURCE_VARS_DICT.
    source_wt_default : int, optional
        The default weight to assign to an edge if no other weight attribute is found. Default is 1.

    Returns
    -------
    pd.DataFrame
        The edges dataframe with the source weights added.
    """

    logger.warning(
        "_create_source_weights should be reimplemented once https://github.com/calico/pathadex-data/issues/95 "
        "is fixed. The current implementation is quite limited."
    )

    # currently, we will look for values of source_indicator_var which are non NA and set them to
    # source_indicator_match_score and setting entries which are NA as source_indicator_nonmatch_score.
    #
    # this is a simple way of flagging string vs. non-string scores

    included_weight_vars = set(source_vars_dict.keys()).intersection(
        set(edges_df.columns)
    )
    if len(included_weight_vars) == 0:
        logger.warning(
            f"No edge attributes were found which match those in source_vars_dict: {', '.join(source_vars_dict.keys())}"
        )
        edges_df[source_wt_var] = source_wt_default
        return edges_df

    edges_df_source_wts = edges_df[list(included_weight_vars)].copy()
    for wt in list(included_weight_vars):
        edges_df_source_wts[wt] = [
            source_wt_default if x is True else source_vars_dict[wt]
            for x in edges_df[wt].isna()
        ]

    source_wt_edges_df = edges_df.join(
        edges_df_source_wts.max(axis=1).rename(source_wt_var)
    )

    return source_wt_edges_df


def _wt_transformation_identity(x):
    """
    Identity transformation for weights.

    Parameters
    ----------
    x : any
        Input value.

    Returns
    -------
    any
        The input value unchanged.
    """
    return x


def _wt_transformation_string(x):
    """
    Map STRING scores to a similar scale as topology weights.

    Parameters
    ----------
    x : float
        STRING score.

    Returns
    -------
    float
        Transformed STRING score.
    """
    return 250000 / np.power(x, 1.7)


def _wt_transformation_string_inv(x):
    """
    Map STRING scores so they work with source weights.

    Parameters
    ----------
    x : float
        STRING score.

    Returns
    -------
    float
        Inverse transformed STRING score.
    """
    # string scores are bounded on [0, 1000]
    # and score/1000 is roughly a probability that
    # there is a real interaction (physical, genetic, ...)
    # reported string scores are currently on [150, 1000]
    # so this transformation will map these onto {6.67, 1}
    return 1 / (x / 1000)


def _augment_network_nodes(
    network_nodes: pd.DataFrame,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    species_graph_attrs: dict = dict(),
    custom_transformations: Optional[dict] = None,
) -> pd.DataFrame:
    """
    Add species-level attributes, expand network_nodes with s_id and c_id and then map to species-level attributes by s_id.

    This function merges species-level attributes from sbml_dfs into the provided network_nodes DataFrame,
    using the mapping in species_graph_attrs. Optionally, custom transformation functions can be provided
    to transform the attributes as they are added.

    Parameters
    ----------
    network_nodes : pd.DataFrame
        DataFrame of network nodes. Must include columns 'name', 'node_name', and 'node_type'.
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The SBML_dfs object containing species data.
    species_graph_attrs : dict, optional
        Dictionary specifying which attributes to pull from species_data and how to transform them.
        The structure should be {attribute_name: {"table": ..., "variable": ..., "trans": ...}}.
    custom_transformations : dict, optional
        Dictionary mapping transformation names to functions. If provided, these will be checked
        before built-in transformations. Example: {"square": lambda x: x**2}

    Returns
    -------
    pd.DataFrame
        The input network_nodes DataFrame with additional columns for each extracted and transformed attribute.

    Raises
    ------
    ValueError
        If required attributes are missing from network_nodes.
    """
    REQUIRED_NETWORK_NODE_ATTRS = {
        "name",
        "node_name",
        "node_type",
    }

    missing_required_network_nodes_attrs = REQUIRED_NETWORK_NODE_ATTRS.difference(
        set(network_nodes.columns.tolist())
    )
    if len(missing_required_network_nodes_attrs) > 0:
        raise ValueError(
            f"{len(missing_required_network_nodes_attrs)} required attributes were missing "
            "from network_nodes: "
            f"{', '.join(missing_required_network_nodes_attrs)}"
        )

    # include matching s_ids and c_ids of sc_ids
    network_nodes_sid = utils._merge_and_log_overwrites(
        network_nodes,
        sbml_dfs.compartmentalized_species[["s_id", "c_id"]],
        "network nodes",
        left_on="name",
        right_index=True,
        how="left",
    )

    # assign species_data related attributes to s_id
    species_graph_data = pluck_entity_data(
        sbml_dfs,
        species_graph_attrs,
        "species",
        custom_transformations=custom_transformations,
    )

    if species_graph_data is not None:
        # add species_graph_data to the network_nodes df, based on s_id
        network_nodes_wdata = utils._merge_and_log_overwrites(
            network_nodes_sid,
            species_graph_data,
            "species graph data",
            left_on="s_id",
            right_index=True,
            how="left",
        )
    else:
        network_nodes_wdata = network_nodes_sid

    # Note: multiple sc_ids with the same s_id will be assign with the same species_graph_data

    network_nodes_wdata = network_nodes_wdata.fillna(int(0)).drop(
        columns=["s_id", "c_id"]
    )

    return network_nodes_wdata


def _augment_network_edges(
    network_edges: pd.DataFrame,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    reaction_graph_attrs: dict = dict(),
    custom_transformations: Optional[dict] = None,
) -> pd.DataFrame:
    """
    Add reversibility and other metadata from reactions.

    Parameters
    ----------
    network_edges : pd.DataFrame
        DataFrame of network edges.
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The SBML_dfs object containing reaction data.
    reaction_graph_attrs : dict, optional
        Dictionary of reaction attributes to add.
    custom_transformations : dict, optional
        Dictionary of custom transformation functions to use for attribute transformation.

    Returns
    -------
    pd.DataFrame
        DataFrame of network edges with additional metadata.

    Raises
    ------
    ValueError
        If required attributes are missing from network_edges.
    """
    REQUIRED_NETWORK_EDGE_ATTRS = {
        "from",
        "to",
        "stoichiometry",
        "sbo_term",
        "sc_degree",
        "sc_children",
        "sc_parents",
        "species_type",
        "r_id",
    }

    missing_required_network_edges_attrs = REQUIRED_NETWORK_EDGE_ATTRS.difference(
        set(network_edges.columns.tolist())
    )
    if len(missing_required_network_edges_attrs) > 0:
        raise ValueError(
            f"{len(missing_required_network_edges_attrs)} required attributes were missing "
            "from network_edges: "
            f"{', '.join(missing_required_network_edges_attrs)}"
        )

    network_edges = (
        network_edges[list(REQUIRED_NETWORK_EDGE_ATTRS)]
        # add reaction-level attributes
        .merge(
            sbml_dfs.reactions[SBML_DFS.R_ISREVERSIBLE],
            left_on=SBML_DFS.R_ID,
            right_index=True,
        )
    )

    # add other attributes based on reactions data
    reaction_graph_data = pluck_entity_data(
        sbml_dfs,
        reaction_graph_attrs,
        SBML_DFS.REACTIONS,
        custom_transformations=custom_transformations,
    )
    if reaction_graph_data is not None:
        network_edges = network_edges.merge(
            reaction_graph_data, left_on=SBML_DFS.R_ID, right_index=True, how="left"
        )

    return network_edges


def _reverse_network_edges(augmented_network_edges: pd.DataFrame) -> pd.DataFrame:
    """
    Flip reversible reactions to derive the reverse reaction.

    Parameters
    ----------
    augmented_network_edges : pd.DataFrame
        DataFrame of network edges with metadata.

    Returns
    -------
    pd.DataFrame
        DataFrame with reversed edges for reversible reactions.

    Raises
    ------
    ValueError
        If required variables are missing or if the transformation fails.
    """

    # validate inputs
    required_vars = {NAPISTU_GRAPH_EDGES.FROM, NAPISTU_GRAPH_EDGES.TO}
    missing_required_vars = required_vars.difference(
        set(augmented_network_edges.columns.tolist())
    )

    if len(missing_required_vars) > 0:
        raise ValueError(
            "augmented_network_edges is missing required variables: "
            f"{', '.join(missing_required_vars)}"
        )

    # Check if direction already exists
    if NAPISTU_GRAPH_EDGES.DIRECTION in augmented_network_edges.columns:
        logger.warning(
            f"{NAPISTU_GRAPH_EDGES.DIRECTION} field already exists in augmented_network_edges. "
            "This is unexpected and may indicate an issue in the graph creation process."
        )

    # select all edges derived from reversible reactions
    reversible_reaction_edges = augmented_network_edges[
        augmented_network_edges[NAPISTU_GRAPH_EDGES.R_ISREVERSIBLE]
    ]

    r_reaction_edges = (
        # ignore edges which start in a regulator or catalyst; even for a reversible reaction it
        # doesn't make sense for a regulator to be impacted by a target
        reversible_reaction_edges[
            ~reversible_reaction_edges[NAPISTU_GRAPH_EDGES.SBO_TERM].isin(
                [
                    MINI_SBO_FROM_NAME[x]
                    for x in SBO_MODIFIER_NAMES.union({SBOTERM_NAMES.CATALYST})
                ]
            )
        ]
        # flip parent and child attributes
        .rename(
            {
                NAPISTU_GRAPH_EDGES.FROM: NAPISTU_GRAPH_EDGES.TO,
                NAPISTU_GRAPH_EDGES.TO: NAPISTU_GRAPH_EDGES.FROM,
                NAPISTU_GRAPH_EDGES.SC_CHILDREN: NAPISTU_GRAPH_EDGES.SC_PARENTS,
                NAPISTU_GRAPH_EDGES.SC_PARENTS: NAPISTU_GRAPH_EDGES.SC_CHILDREN,
            },
            axis=1,
        )
    )

    # switch substrates and products
    r_reaction_edges[NAPISTU_GRAPH_EDGES.STOICHIOMETRY] = r_reaction_edges[
        NAPISTU_GRAPH_EDGES.STOICHIOMETRY
    ].apply(
        # the ifelse statement prevents 0 being converted to -0 ...
        lambda x: -1 * x if x != 0 else 0
    )

    transformed_r_reaction_edges = pd.concat(
        [
            (
                r_reaction_edges[
                    r_reaction_edges[NAPISTU_GRAPH_EDGES.SBO_TERM]
                    == MINI_SBO_FROM_NAME[SBOTERM_NAMES.REACTANT]
                ].assign(sbo_term=MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT])
            ),
            (
                r_reaction_edges[
                    r_reaction_edges[NAPISTU_GRAPH_EDGES.SBO_TERM]
                    == MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT]
                ].assign(sbo_term=MINI_SBO_FROM_NAME[SBOTERM_NAMES.REACTANT])
            ),
            r_reaction_edges[
                ~r_reaction_edges[NAPISTU_GRAPH_EDGES.SBO_TERM].isin(
                    [
                        MINI_SBO_FROM_NAME[SBOTERM_NAMES.REACTANT],
                        MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT],
                    ]
                )
            ],
        ]
    )

    if transformed_r_reaction_edges.shape[0] != r_reaction_edges.shape[0]:
        raise ValueError(
            "transformed_r_reaction_edges and r_reaction_edges must have the same number of rows"
        )

    return transformed_r_reaction_edges.assign(
        **{NAPISTU_GRAPH_EDGES.DIRECTION: NAPISTU_GRAPH_EDGE_DIRECTIONS.REVERSE}
    )


def _create_topology_weights(
    napistu_graph: ig.Graph,
    base_score: float = 2,
    protein_multiplier: int = 1,
    metabolite_multiplier: int = 3,
    unknown_multiplier: int = 10,
    scale_multiplier_by_meandegree: bool = True,
) -> ig.Graph:
    """
    Create Topology Weights for a network based on its topology.

    Edges downstream of nodes with many connections receive a higher weight suggesting that any one
    of them is less likely to be regulatory. This is a simple and clearly flawed heuristic which can be
    combined with more principled weighting schemes.

    Parameters
    ----------
    napistu_graph : ig.Graph
        A graph containing connections between molecules, proteins, and reactions.
    base_score : float, optional
        Offset which will be added to all weights. Default is 2.
    protein_multiplier : int, optional
        Multiplier for non-metabolite species. Default is 1.
    metabolite_multiplier : int, optional
        Multiplier for metabolites. Default is 3.
    unknown_multiplier : int, optional
        Multiplier for species without any identifier. Default is 10.
    scale_multiplier_by_meandegree : bool, optional
        If True, multipliers will be rescaled by the average number of connections a node has. Default is True.

    Returns
    -------
    ig.Graph
        Graph with added topology weights.

    Raises
    ------
    ValueError
        If required attributes are missing or if parameters are invalid.
    """

    # check for required attribute before proceeding

    required_attrs = {
        NAPISTU_GRAPH_EDGES.SC_DEGREE,
        NAPISTU_GRAPH_EDGES.SC_CHILDREN,
        NAPISTU_GRAPH_EDGES.SC_PARENTS,
        NAPISTU_GRAPH_EDGES.SPECIES_TYPE,
    }

    missing_required_attrs = required_attrs.difference(
        set(napistu_graph.es.attributes())
    )
    if len(missing_required_attrs) != 0:
        raise ValueError(
            f"model is missing {len(missing_required_attrs)} required attributes: {', '.join(missing_required_attrs)}"
        )

    if base_score < 0:
        raise ValueError(f"base_score was {base_score} and must be non-negative")
    if protein_multiplier > unknown_multiplier:
        raise ValueError(
            f"protein_multiplier was {protein_multiplier} and unknown_multiplier "
            f"was {unknown_multiplier}. unknown_multiplier must be greater than "
            "protein_multiplier"
        )
    if metabolite_multiplier > unknown_multiplier:
        raise ValueError(
            f"protein_multiplier was {metabolite_multiplier} and unknown_multiplier "
            f"was {unknown_multiplier}. unknown_multiplier must be greater than "
            "protein_multiplier"
        )

    # create a new weight variable

    weight_table = pd.DataFrame(
        {
            NAPISTU_GRAPH_EDGES.SC_DEGREE: napistu_graph.es[
                NAPISTU_GRAPH_EDGES.SC_DEGREE
            ],
            NAPISTU_GRAPH_EDGES.SC_CHILDREN: napistu_graph.es[
                NAPISTU_GRAPH_EDGES.SC_CHILDREN
            ],
            NAPISTU_GRAPH_EDGES.SC_PARENTS: napistu_graph.es[
                NAPISTU_GRAPH_EDGES.SC_PARENTS
            ],
            NAPISTU_GRAPH_EDGES.SPECIES_TYPE: napistu_graph.es[
                NAPISTU_GRAPH_EDGES.SPECIES_TYPE
            ],
        }
    )

    lookup_multiplier_dict = {
        "protein": protein_multiplier,
        "metabolite": metabolite_multiplier,
        "unknown": unknown_multiplier,
    }
    weight_table["multiplier"] = weight_table["species_type"].map(
        lookup_multiplier_dict
    )

    # calculate mean degree
    # since topology weights will differ based on the structure of the network
    # and it would be nice to have a consistent notion of edge weights and path weights
    # for interpretability and filtering, we can rescale topology weights by the
    # average degree of nodes
    if scale_multiplier_by_meandegree:
        mean_degree = len(napistu_graph.es) / len(napistu_graph.vs)
        if not napistu_graph.is_directed():
            # for a directed network in- and out-degree are separately treated while
            # an undirected network's degree will be the sum of these two measures.
            mean_degree = mean_degree * 2

        weight_table["multiplier"] = weight_table["multiplier"] / mean_degree

    if napistu_graph.is_directed():
        weight_table["connection_weight"] = weight_table[
            NAPISTU_GRAPH_EDGES.SC_CHILDREN
        ]
    else:
        weight_table["connection_weight"] = weight_table[NAPISTU_GRAPH_EDGES.SC_DEGREE]

    # weight traveling through a species based on
    # - a constant
    # - how plausibly that species type mediates a change
    # - the number of connections that the node can bridge to
    weight_table["topo_weights"] = [
        base_score + (x * y)
        for x, y in zip(weight_table["multiplier"], weight_table["connection_weight"])
    ]
    napistu_graph.es["topo_weights"] = weight_table["topo_weights"]

    # if directed and we want to use travel upstream define a corresponding weighting scheme
    if napistu_graph.is_directed():
        weight_table["upstream_topo_weights"] = [
            base_score + (x * y)
            for x, y in zip(weight_table["multiplier"], weight_table["sc_parents"])
        ]
        napistu_graph.es["upstream_topo_weights"] = weight_table[
            "upstream_topo_weights"
        ]

    return napistu_graph


def _validate_entity_attrs(
    entity_attrs: dict,
    validate_transformations: bool = True,
    custom_transformations: Optional[dict] = None,
) -> None:
    """
    Validate that graph attributes are a valid format.

    Parameters
    ----------
    entity_attrs : dict
        Dictionary of entity attributes to validate.
    validate_transformations : bool, optional
        Whether to validate transformation names, by default True.
    custom_transformations : dict, optional
        Dictionary of custom transformation functions, by default None. Keys are transformation names, values are transformation functions.

    Returns
    -------
    None

    Raises
    ------
    AssertionError
        If entity_attrs is not a dictionary.
    ValueError
        If a transformation is not found in DEFINED_WEIGHT_TRANSFORMATION or custom_transformations.
    """
    assert isinstance(entity_attrs, dict), "entity_attrs must be a dictionary"

    for k, v in entity_attrs.items():
        # check structure against pydantic config
        validated_attrs = _EntityAttrValidator(**v).model_dump()

        if validate_transformations:
            trans_name = validated_attrs.get("trans", DEFAULT_WT_TRANS)
            valid_trans = set(DEFINED_WEIGHT_TRANSFORMATION.keys())
            if custom_transformations:
                valid_trans = valid_trans.union(set(custom_transformations.keys()))
            if trans_name not in valid_trans:
                raise ValueError(
                    f"transformation '{trans_name}' was not defined as an alias in "
                    "DEFINED_WEIGHT_TRANSFORMATION or custom_transformations. The defined transformations "
                    f"are {', '.join(sorted(valid_trans))}"
                )

    return None


class _EntityAttrValidator(BaseModel):
    table: str
    variable: str
    trans: Optional[str] = DEFAULT_WT_TRANS



"""
Utilities for scanning and analyzing the Napistu codebase.
"""

from typing import Dict, Optional, Any, Set

from napistu.mcp import utils as mcp_utils
from napistu.mcp.constants import READTHEDOCS_TOC_CSS_SELECTOR

# Import optional dependencies with error handling
try:
    from bs4 import BeautifulSoup
except ImportError:
    raise ImportError(
        "Documentation utilities require additional dependencies. Install with 'pip install napistu[mcp]'"
    )


async def read_read_the_docs(package_toc_url: str) -> dict:
    """
    Recursively parse all modules and submodules starting from the package TOC.
    """
    # Step 1: Get all module URLs from the TOC
    packages_dict = await _process_rtd_package_toc(package_toc_url)
    docs_dict = {}
    visited = set()

    # Step 2: Recursively parse each module page
    for package_name, module_url in packages_dict.items():
        if not module_url.startswith("http"):
            # Make absolute if needed
            base = package_toc_url.rsplit("/", 1)[0]
            module_url = base + "/" + module_url.lstrip("/")
        await _parse_rtd_module_recursive(module_url, visited, docs_dict)

    return docs_dict


def extract_functions_and_classes_from_modules(modules: dict) -> tuple[dict, dict]:
    """
    Process the modules cache and return a tuple (functions_dict, classes_dict),
    where each is a dict keyed by fully qualified name (e.g., 'module.func', 'module.Class').
    Recursively processes submodules.

    Args:
        modules (dict): The modules cache as returned by read_read_the_docs.

    Returns:
        tuple: (functions_dict, classes_dict)
    """
    functions = {}
    classes = {}

    def _process_module(module_name: str, module_info: dict):
        # Functions
        for func_name, func_info in module_info.get("functions", {}).items():
            fq_name = f"{module_name}.{func_name}"
            functions[fq_name] = func_info
        # Classes
        for class_name, class_info in module_info.get("classes", {}).items():
            fq_name = f"{module_name}.{class_name}"
            classes[fq_name] = class_info
        # Submodules (if present in the cache)
        for submod_name in module_info.get("submodules", {}):
            fq_submod_name = f"{module_name}.{submod_name}"
            if fq_submod_name in modules:
                _process_module(fq_submod_name, modules[fq_submod_name])

    for module_name, module_info in modules.items():
        _process_module(module_name, module_info)

    return functions, classes


def _parse_rtd_module_page(html: str, url: Optional[str] = None) -> dict:
    """
    Parse a ReadTheDocs module HTML page and extract functions, classes, methods, attributes, and submodules.
    Returns a dict suitable for MCP server use, with functions, classes, and methods keyed by name.

    Args:
        html (str): The HTML content of the module page.
        url (Optional[str]): The URL of the page (for reference).

    Returns:
        dict: {
            'module': str,
            'url': str,
            'functions': Dict[str, dict],
            'classes': Dict[str, dict],
            'submodules': Dict[str, dict]
        }
    """
    soup = BeautifulSoup(html, "html.parser")
    result = {
        "module": None,
        "url": url,
        "functions": {},
        "classes": {},
        "submodules": _format_submodules(soup),
    }
    # Get module name from <h1>
    h1 = soup.find("h1")
    if h1:
        module_name = h1.get_text(strip=True).replace("\uf0c1", "").strip()
        result["module"] = module_name
    # Functions
    for func_dl in soup.find_all("dl", class_="py function"):
        func = _format_function(func_dl.find("dt"), func_dl.find("dd"))
        if func["name"]:
            result["functions"][func["name"]] = func
    # Classes
    for class_dl in soup.find_all("dl", class_="py class"):
        cls = _format_class(class_dl)
        if cls["name"]:
            result["classes"][cls["name"]] = cls
    return result


async def _process_rtd_package_toc(
    url: str, css_selector: str = READTHEDOCS_TOC_CSS_SELECTOR
) -> dict:
    """
    Parse the ReadTheDocs package TOC and return a dict of {name: url}.
    """
    page_html = await mcp_utils.load_html_page(url)
    soup = BeautifulSoup(page_html, "html.parser")
    selected = soup.select(css_selector)
    return _parse_module_tags(selected)


def _parse_module_tags(td_list: list, base_url: str = "") -> dict:
    """
    Parse a list of <td> elements containing module links and return a dict of {name: url}.
    Optionally prepends base_url to relative hrefs.
    """
    result = {}
    for td in td_list:
        a = td.find("a", class_="reference internal")
        if a:
            # Get the module name from the <span class="pre"> tag
            span = a.find("span", class_="pre")
            if span:
                name = span.text.strip()
                href = a.get("href")
                # Prepend base_url if href is relative
                if href and not href.startswith("http"):
                    href = base_url.rstrip("/") + "/" + href.lstrip("/")
                result[name] = href
    return result


def _format_function(sig_dt, doc_dd) -> Dict[str, Any]:
    """
    Format a function or method signature and its documentation into a dictionary.

    Args:
        sig_dt: The <dt> tag containing the function/method signature.
        doc_dd: The <dd> tag containing the function/method docstring.

    Returns:
        dict: A dictionary with keys 'name', 'signature', 'id', and 'doc'.
    """
    name = (
        sig_dt.find("span", class_="sig-name").get_text(strip=True) if sig_dt else None
    )
    signature = sig_dt.get_text(strip=True) if sig_dt else None
    return {
        "name": mcp_utils._clean_signature_text(name),
        "signature": mcp_utils._clean_signature_text(signature),
        "id": sig_dt.get("id") if sig_dt else None,
        "doc": doc_dd.get_text(" ", strip=True) if doc_dd else None,
    }


def _format_attribute(attr_dl) -> Dict[str, Any]:
    """
    Format a class attribute's signature and documentation into a dictionary.

    Args:
        attr_dl: The <dl> tag for the attribute, containing <dt> and <dd>.

    Returns:
        dict: A dictionary with keys 'name', 'signature', 'id', and 'doc'.
    """
    sig = attr_dl.find("dt")
    doc = attr_dl.find("dd")
    name = sig.find("span", class_="sig-name").get_text(strip=True) if sig else None
    signature = sig.get_text(strip=True) if sig else None
    return {
        "name": mcp_utils._clean_signature_text(name),
        "signature": mcp_utils._clean_signature_text(signature),
        "id": sig.get("id") if sig else None,
        "doc": doc.get_text(" ", strip=True) if doc else None,
    }


def _format_class(class_dl) -> Dict[str, Any]:
    """
    Format a class definition, including its methods and attributes, into a dictionary.

    Args:
        class_dl: The <dl> tag for the class, containing <dt> and <dd>.

    Returns:
        dict: A dictionary with keys 'name', 'signature', 'id', 'doc', 'methods', and 'attributes'.
              'methods' and 'attributes' are themselves dicts keyed by name.
    """
    sig = class_dl.find("dt")
    doc = class_dl.find("dd")
    class_name = (
        sig.find("span", class_="sig-name").get_text(strip=True) if sig else None
    )
    methods = {}
    attributes = {}
    if doc:
        for meth_dl in doc.find_all("dl", class_="py method"):
            meth = _format_function(meth_dl.find("dt"), meth_dl.find("dd"))
            if meth["name"]:
                methods[meth["name"]] = meth
        for attr_dl in doc.find_all("dl", class_="py attribute"):
            attr = _format_attribute(attr_dl)
            if attr["name"]:
                attributes[attr["name"]] = attr
    return {
        "name": mcp_utils._clean_signature_text(class_name),
        "signature": mcp_utils._clean_signature_text(
            sig.get_text(strip=True) if sig else None
        ),
        "id": sig.get("id") if sig else None,
        "doc": doc.get_text(" ", strip=True) if doc else None,
        "methods": methods,
        "attributes": attributes,
    }


def _format_submodules(soup) -> dict:
    """
    Extract submodules from a ReadTheDocs module page soup object.
    Looks for a 'Modules' rubric and parses the following table or list for submodule names, URLs, and descriptions.

    Args:
        soup (BeautifulSoup): Parsed HTML soup of the module page.

    Returns:
        dict: {submodule_name: {"url": str, "description": str}}
    """
    submodules = {}
    for rubric in soup.find_all("p", class_="rubric"):
        if rubric.get_text(strip=True).lower() == "modules":
            sib = rubric.find_next_sibling()
            if sib and sib.name in ("table", "ul"):
                for a in sib.find_all("a", href=True):
                    submod_name = a.get_text(strip=True)
                    submod_url = a["href"]
                    desc = ""
                    td = a.find_parent("td")
                    if td and td.find_next_sibling("td"):
                        desc = td.find_next_sibling("td").get_text(strip=True)
                    elif a.parent.name == "li":
                        next_p = a.find_next_sibling("p")
                        if next_p:
                            desc = next_p.get_text(strip=True)
                    submodules[submod_name] = {"url": submod_url, "description": desc}
    return submodules


async def _parse_rtd_module_recursive(
    module_url: str,
    visited: Optional[Set[str]] = None,
    docs_dict: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """
    Recursively parse a module page and all its submodules.
    """

    if visited is None:
        visited = set()
    if docs_dict is None:
        docs_dict = {}

    if module_url in visited:
        return docs_dict
    visited.add(module_url)

    page_html = await mcp_utils.load_html_page(module_url)
    module_doc = _parse_rtd_module_page(page_html, module_url)
    module_name = module_doc.get("module") or module_url
    docs_dict[module_name] = module_doc

    # Recursively parse submodules
    for submod_name, submod_info in module_doc.get("submodules", {}).items():
        submod_url = submod_info["url"]
        if not submod_url.startswith("http"):
            base = module_url.rsplit("/", 1)[0]
            submod_url = base + "/" + submod_url.lstrip("/")
        await _parse_rtd_module_recursive(submod_url, visited, docs_dict)

    return docs_dict



from typing import Union

import numpy as np
from scipy.stats import norm


def fisher_exact_vectorized(
    observed_members: Union[list[int], np.ndarray],
    missing_members: Union[list[int], np.ndarray],
    observed_nonmembers: Union[list[int], np.ndarray],
    nonobserved_nonmembers: Union[list[int], np.ndarray],
) -> tuple[np.ndarray, np.ndarray]:
    """
    Fast vectorized one-tailed Fisher exact test using normal approximation.

    Parameters:
    -----------
    observed_members, missing_members, observed_nonmembers, nonobserved_nonmembers : array-like
        The four cells of the 2x2 contingency tables (must be non-negative)

    Returns:
    --------
    odds_ratios : numpy array
        Odds ratios for each test
    p_values : numpy array
        One-tailed p-values (tests for enrichment)
    """
    # Convert to numpy arrays
    a = np.array(observed_members, dtype=float)
    b = np.array(missing_members, dtype=float)
    c = np.array(observed_nonmembers, dtype=float)
    d = np.array(nonobserved_nonmembers, dtype=float)

    # Check for negative values and raise error
    if np.any((a < 0) | (b < 0) | (c < 0) | (d < 0)):
        raise ValueError("All contingency table values must be non-negative")

    # Calculate odds ratios
    odds_ratios = np.divide(
        a * d, b * c, out=np.full_like(a, np.inf, dtype=float), where=(b * c) != 0
    )

    # Normal approximation to hypergeometric distribution
    n = a + b + c + d

    # Avoid division by zero in expected value calculation
    expected_a = np.divide(
        (a + b) * (a + c), n, out=np.zeros_like(n, dtype=float), where=n != 0
    )

    # Variance calculation with protection against division by zero
    var_a = np.divide(
        (a + b) * (c + d) * (a + c) * (b + d),
        n * n * (n - 1),
        out=np.ones_like(n, dtype=float),  # Default to 1 to avoid sqrt(0)
        where=(n > 1),
    )
    var_a = np.maximum(var_a, 1e-10)  # Ensure positive variance

    # Continuity correction and z-score
    z = (a - expected_a - 0.5) / np.sqrt(var_a)

    # One-tailed p-value (upper tail for enrichment)
    p_values = norm.sf(z)  # 1 - norm.cdf(z)

    return odds_ratios, p_values



from __future__ import annotations

import sys
from unittest.mock import Mock, patch
import pytest


# Mock all rpy2 dependencies before any imports to prevent ImportErrors
sys.modules["rpy2"] = Mock()
sys.modules["rpy2.robjects"] = Mock()
sys.modules["rpy2.robjects.conversion"] = Mock()
sys.modules["rpy2.robjects.default_converter"] = Mock()
sys.modules["rpy2.robjects.packages"] = Mock()
sys.modules["rpy2.robjects.pandas2ri"] = Mock()
sys.modules["rpy2.rinterface"] = Mock()
sys.modules["rpy2_arrow"] = Mock()
sys.modules["rpy2_arrow.arrow"] = Mock()

import napistu.rpy2  # noqa: E402


def test_rpy2_availability_detection():
    """Test rpy2 availability detection in various scenarios."""
    # Test ImportError case
    with patch.dict("sys.modules", {"rpy2": None}):
        if hasattr(napistu.rpy2.get_rpy2_availability, "cache_clear"):
            napistu.rpy2.get_rpy2_availability.cache_clear()
        assert napistu.rpy2.get_rpy2_availability() is False

    # Test other exception case during import
    with patch("builtins.__import__") as mock_import:
        mock_import.side_effect = RuntimeError("R installation broken")
        if hasattr(napistu.rpy2.get_rpy2_availability, "cache_clear"):
            napistu.rpy2.get_rpy2_availability.cache_clear()
        assert napistu.rpy2.get_rpy2_availability() is False

    # Test success case
    mock_rpy2 = Mock()
    with patch.dict("sys.modules", {"rpy2": mock_rpy2}):
        if hasattr(napistu.rpy2.get_rpy2_availability, "cache_clear"):
            napistu.rpy2.get_rpy2_availability.cache_clear()
        assert napistu.rpy2.get_rpy2_availability() is True


def test_caching_behavior():
    """Test that lazy loading functions are properly cached."""
    # Test availability caching
    with patch("builtins.__import__") as mock_import:
        mock_import.return_value = Mock()
        if hasattr(napistu.rpy2.get_rpy2_availability, "cache_clear"):
            napistu.rpy2.get_rpy2_availability.cache_clear()

        result1 = napistu.rpy2.get_rpy2_availability()
        result2 = napistu.rpy2.get_rpy2_availability()

        assert result1 == result2
        # Only test call count if caching is enabled
        if hasattr(napistu.rpy2.get_rpy2_availability, "cache_clear"):
            assert mock_import.call_count == 1  # Should only be called once

    # Test core modules caching
    with patch("napistu.rpy2.get_rpy2_availability", return_value=True):
        with patch("rpy2.robjects.conversion"), patch(
            "rpy2.robjects.default_converter"
        ), patch("rpy2.robjects.packages.importr"):
            if hasattr(napistu.rpy2.get_rpy2_core_modules, "cache_clear"):
                napistu.rpy2.get_rpy2_core_modules.cache_clear()

            result1 = napistu.rpy2.get_rpy2_core_modules()
            result2 = napistu.rpy2.get_rpy2_core_modules()

            # Only test object identity if caching is enabled
            if hasattr(napistu.rpy2.get_rpy2_core_modules, "cache_clear"):
                assert result1 is result2  # Same object due to caching


def test_lazy_import_functions_without_rpy2():
    """Test that lazy import functions fail appropriately when rpy2 unavailable."""
    with patch("napistu.rpy2.get_rpy2_availability", return_value=False):
        # Clear all caches if they exist
        if hasattr(napistu.rpy2.get_rpy2_core_modules, "cache_clear"):
            napistu.rpy2.get_rpy2_core_modules.cache_clear()
        if hasattr(napistu.rpy2.get_rpy2_extended_modules, "cache_clear"):
            napistu.rpy2.get_rpy2_extended_modules.cache_clear()
        if hasattr(napistu.rpy2.get_napistu_r_package, "cache_clear"):
            napistu.rpy2.get_napistu_r_package.cache_clear()

        # All should raise ImportError
        with pytest.raises(ImportError, match="requires `rpy2`"):
            napistu.rpy2.get_rpy2_core_modules()

        with pytest.raises(ImportError, match="requires `rpy2`"):
            napistu.rpy2.get_rpy2_extended_modules()

        with pytest.raises(ImportError, match="requires `rpy2`"):
            napistu.rpy2.get_napistu_r_package()


def test_decorators():
    """Test require_rpy2 and report_r_exceptions decorators."""
    # Test require_rpy2 with rpy2 available
    with patch("napistu.rpy2.get_rpy2_availability", return_value=True):

        @napistu.rpy2.require_rpy2
        def test_func_success():
            return "success"

        assert test_func_success() == "success"

    # Test require_rpy2 without rpy2
    with patch("napistu.rpy2.get_rpy2_availability", return_value=False):

        @napistu.rpy2.require_rpy2
        def test_func_fail():
            return "success"

        with pytest.raises(ImportError, match="test_func_fail.*requires `rpy2`"):
            test_func_fail()

    # Test report_r_exceptions with success
    with patch("napistu.rpy2.get_rpy2_availability", return_value=True):

        @napistu.rpy2.report_r_exceptions
        def test_func_report_success():
            return "success"

        assert test_func_report_success() == "success"

    # Test report_r_exceptions with failure
    with patch("napistu.rpy2.get_rpy2_availability", return_value=True):
        with patch("napistu.rpy2.rsession_info") as mock_rsession:

            @napistu.rpy2.report_r_exceptions
            def test_func_report_fail():
                raise ValueError("R function failed")

            with pytest.raises(ValueError, match="R function failed"):
                test_func_report_fail()

            mock_rsession.assert_called_once()


def test_module_import_safety():
    """Test that modules can be imported safely without triggering R initialization."""
    import importlib

    # Test modules can be imported without rpy2
    with patch.dict("sys.modules", {"rpy2": None}):
        # These should not raise ImportError during import
        importlib.reload(napistu.rpy2)



from __future__ import annotations

import os

import numpy as np
import pandas as pd
from napistu import identifiers
from napistu.constants import IDENTIFIERS
import pytest

# logger = logging.getLogger()
# logger.setLevel("DEBUG")

test_path = os.path.abspath(os.path.join(__file__, os.pardir))
identifier_examples = pd.read_csv(
    os.path.join(test_path, "test_data", "identifier_examples.tsv"),
    sep="\t",
    header=0,
)


def test_identifiers():
    assert (
        identifiers.Identifiers(
            [{"ontology": "KEGG", "identifier": "C00031", "bqb": "BQB_IS"}]
        ).ids[0]["ontology"]
        == "KEGG"
    )

    example_identifiers = identifiers.Identifiers(
        [
            {"ontology": "SGD", "identifier": "S000004535", "bqb": "BQB_IS"},
            {"ontology": "foo", "identifier": "bar", "bqb": "BQB_IS"},
        ]
    )

    assert type(example_identifiers) is identifiers.Identifiers

    assert example_identifiers.filter("SGD") is True
    assert example_identifiers.filter("baz") is False
    assert example_identifiers.filter("SGD", summarize=False) == [True, False]
    assert example_identifiers.filter(["SGD", "foo"], summarize=False) == [True, True]
    assert example_identifiers.filter(["foo", "SGD"], summarize=False) == [True, True]
    assert example_identifiers.filter(["baz", "bar"], summarize=False) == [False, False]

    assert example_identifiers.hoist("SGD") == "S000004535"
    assert example_identifiers.hoist("baz") is None


def test_identifiers_from_urls():
    for i in range(0, identifier_examples.shape[0]):
        # print(identifier_examples["url"][i])
        testIdentifiers = identifiers.Identifiers(
            [
                identifiers.format_uri(
                    identifier_examples["url"][i], biological_qualifier_type="BQB_IS"
                )
            ]
        )

        # print(f"ontology = {testIdentifiers.ids[0]['ontology']}; identifier = {testIdentifiers.ids[0]['identifier']}")
        assert (
            testIdentifiers.ids[0]["ontology"] == identifier_examples["ontology"][i]
        ), f"ontology {testIdentifiers.ids[0]['ontology']} does not equal {identifier_examples['ontology'][i]}"

        assert (
            testIdentifiers.ids[0]["identifier"] == identifier_examples["identifier"][i]
        ), f"identifier {testIdentifiers.ids[0]['identifier']} does not equal {identifier_examples['identifier'][i]}"


def test_url_from_identifiers():
    for row in identifier_examples.iterrows():
        # some urls (e.g., chebi) will be converted to a canonical url (e.g., chebi) since multiple URIs exist

        if row[1]["canonical_url"] is not np.nan:
            expected_url_out = row[1]["canonical_url"]
        else:
            expected_url_out = row[1]["url"]

        url_out = identifiers.create_uri_url(
            ontology=row[1]["ontology"], identifier=row[1]["identifier"]
        )

        # print(f"expected: {expected_url_out}; observed: {url_out}")
        assert url_out == expected_url_out

    # test non-strict treatment

    assert (
        identifiers.create_uri_url(ontology="chebi", identifier="abc", strict=False)
        is None
    )


def test_parsing_ensembl_ids():
    ensembl_examples = {
        # human foxp2
        "ENSG00000128573": ("ENSG00000128573", "ensembl_gene", "Homo sapiens"),
        "ENST00000441290": ("ENST00000441290", "ensembl_transcript", "Homo sapiens"),
        "ENSP00000265436": ("ENSP00000265436", "ensembl_protein", "Homo sapiens"),
        # mouse leptin
        "ENSMUSG00000059201": ("ENSMUSG00000059201", "ensembl_gene", "Mus musculus"),
        "ENSMUST00000069789": (
            "ENSMUST00000069789",
            "ensembl_transcript",
            "Mus musculus",
        ),
        # substrings are okay
        "gene=ENSMUSG00000017146": (
            "ENSMUSG00000017146",
            "ensembl_gene",
            "Mus musculus",
        ),
    }

    for k, v in ensembl_examples.items():
        assert identifiers.parse_ensembl_id(k) == v


def test_reciprocal_ensembl_dicts():
    assert len(identifiers.ENSEMBL_SPECIES_TO_CODE) == len(
        identifiers.ENSEMBL_SPECIES_FROM_CODE
    )
    for k in identifiers.ENSEMBL_SPECIES_TO_CODE.keys():
        assert (
            identifiers.ENSEMBL_SPECIES_FROM_CODE[
                identifiers.ENSEMBL_SPECIES_TO_CODE[k]
            ]
            == k
        )

    assert len(identifiers.ENSEMBL_MOLECULE_TYPES_TO_ONTOLOGY) == len(
        identifiers.ENSEMBL_MOLECULE_TYPES_FROM_ONTOLOGY
    )
    for k in identifiers.ENSEMBL_MOLECULE_TYPES_TO_ONTOLOGY.keys():
        assert (
            identifiers.ENSEMBL_MOLECULE_TYPES_FROM_ONTOLOGY[
                identifiers.ENSEMBL_MOLECULE_TYPES_TO_ONTOLOGY[k]
            ]
            == k
        )


def test_df_to_identifiers_basic():
    """Test basic conversion of DataFrame to Identifiers objects."""
    # Create a simple test DataFrame
    df = pd.DataFrame(
        {
            "s_id": ["s1", "s1", "s2"],
            IDENTIFIERS.ONTOLOGY: ["ncbi_entrez_gene", "uniprot", "ncbi_entrez_gene"],
            IDENTIFIERS.IDENTIFIER: ["123", "P12345", "456"],
            IDENTIFIERS.URL: [
                "http://ncbi/123",
                "http://uniprot/P12345",
                "http://ncbi/456",
            ],
            IDENTIFIERS.BQB: ["is", "is", "is"],
        }
    )

    # Convert to Identifiers objects
    result = identifiers.df_to_identifiers(df)

    # Check basic properties
    assert isinstance(result, pd.Series)
    assert len(result) == 2  # Two unique s_ids
    assert all(isinstance(x, identifiers.Identifiers) for x in result)

    # Check specific values
    s1_ids = result["s1"].ids
    assert len(s1_ids) == 2  # Two identifiers for s1
    assert any(x[IDENTIFIERS.IDENTIFIER] == "123" for x in s1_ids)
    assert any(x[IDENTIFIERS.IDENTIFIER] == "P12345" for x in s1_ids)

    s2_ids = result["s2"].ids
    assert len(s2_ids) == 1  # One identifier for s2
    assert s2_ids[0][IDENTIFIERS.IDENTIFIER] == "456"


def test_df_to_identifiers_duplicates():
    """Test that duplicates are handled correctly."""
    # Create DataFrame with duplicate entries
    df = pd.DataFrame(
        {
            "s_id": ["s1", "s1", "s1"],
            IDENTIFIERS.ONTOLOGY: [
                "ncbi_entrez_gene",
                "ncbi_entrez_gene",
                "ncbi_entrez_gene",
            ],
            IDENTIFIERS.IDENTIFIER: ["123", "123", "123"],  # Same identifier repeated
            IDENTIFIERS.URL: ["http://ncbi/123"] * 3,
            IDENTIFIERS.BQB: ["is"] * 3,
        }
    )

    result = identifiers.df_to_identifiers(df)

    # Should collapse duplicates
    assert len(result) == 1  # One unique s_id
    assert len(result["s1"].ids) == 1  # One unique identifier


def test_df_to_identifiers_missing_columns():
    """Test that missing required columns raise an error."""
    # Create DataFrame missing required columns
    df = pd.DataFrame(
        {
            "s_id": ["s1"],
            IDENTIFIERS.ONTOLOGY: ["ncbi_entrez_gene"],
            IDENTIFIERS.IDENTIFIER: ["123"],
            # Missing URL and BQB
        }
    )

    with pytest.raises(
        ValueError, match="The DataFrame does not contain the required columns"
    ):
        identifiers.df_to_identifiers(df)



"""Module for handling ontology aliases and validation."""

from __future__ import annotations

from typing import Dict, Set
import logging
from pydantic import BaseModel, field_validator
from napistu.constants import (
    ONTOLOGY_SPECIES_ALIASES,
    ONTOLOGIES_LIST,
    IDENTIFIERS,
    SBML_DFS,
)
import pandas as pd
from napistu import identifiers, sbml_dfs_core

logger = logging.getLogger(__name__)


def rename_species_ontologies(
    sbml_dfs: sbml_dfs_core.SBML_dfs, aliases=ONTOLOGY_SPECIES_ALIASES
):
    """
    Rename ontologies in the species identifiers table of an SBML_dfs object using provided aliases.

    This function updates the ontology names in the species identifiers of the given SBML_dfs object
    according to the provided alias mapping. It validates the alias mapping, logs which ontologies will be updated,
    and replaces any matching aliases in the species identifiers with their canonical ontology names.

    Parameters
    ----------
    sbml_dfs : napistu.sbml_dfs_core.SBML_dfs
        The SBML_dfs object whose species table will be updated in-place.
    aliases : dict[str, set[str]], optional
        Dictionary mapping canonical ontology names to sets of their aliases. By default, uses ONTOLOGY_SPECIES_ALIASES.
        All keys must be valid ontologies from ONTOLOGIES_LIST. Values must not overlap between keys or with keys themselves.

    Returns
    -------
    None
        The function updates sbml_dfs.species in-place and does not return a value.

    Raises
    ------
    ValueError
        If the alias mapping is invalid (e.g., keys not in ONTOLOGIES_LIST, overlapping values, or values used as keys),
        or if there is no overlap between the provided aliases and the ontologies present in the species identifiers.

    Examples
    --------
    >>> from napistu.ontologies.renaming import rename_species_ontologies
    >>> sbml_dfs = ...  # an SBML_dfs object
    >>> aliases = {"ncbi_entrez_gene": {"ncbigene", "ncbi_gene"}, "uniprot": {"uniprot_id"}}
    >>> rename_species_ontologies(sbml_dfs, aliases)
    """

    species_identifiers = sbml_dfs.get_identifiers(SBML_DFS.SPECIES)

    aliases = OntologySet(ontologies=aliases).ontologies
    alias_mapping = _create_alias_mapping(aliases)

    _log_ontology_updates(alias_mapping, set(species_identifiers[IDENTIFIERS.ONTOLOGY]))

    species_identifiers[IDENTIFIERS.ONTOLOGY] = species_identifiers[
        IDENTIFIERS.ONTOLOGY
    ].map(lambda x: alias_mapping.get(x, x))

    species_identifiers = identifiers.df_to_identifiers(species_identifiers)

    updated_species = sbml_dfs.species.drop(SBML_DFS.S_IDENTIFIERS, axis=1).join(
        pd.DataFrame(species_identifiers)
    )
    # fill missing attributes with empty Identifiers
    updated_species[SBML_DFS.S_IDENTIFIERS] = updated_species[
        SBML_DFS.S_IDENTIFIERS
    ].fillna(identifiers.Identifiers([]))

    setattr(sbml_dfs, "species", updated_species)


class OntologySet(BaseModel):
    """Validate ontology mappings.

    This model ensures that:
    1. All keys are valid ontologies from ONTOLOGIES_LIST
    2. The dict maps strings to sets of strings
    3. Values in the sets do not overlap between different keys
    4. Values in the sets are not also used as keys

    Attributes
    ----------
    ontologies : Dict[str, Set[str]]
        Dictionary mapping ontology names to sets of their aliases
    """

    ontologies: Dict[str, Set[str]]

    @field_validator("ontologies")
    @classmethod
    def validate_ontologies(cls, v: Dict[str, Set[str]]) -> Dict[str, Set[str]]:
        """Validate the ontology mapping structure.

        Parameters
        ----------
        v : Dict[str, Set[str]]
            Dictionary mapping ontology names to sets of their aliases

        Returns
        -------
        Dict[str, Set[str]]
            The validated ontology mapping dictionary

        Raises
        ------
        ValueError
            If any keys are not valid ontologies from ONTOLOGIES_LIST
            If any values overlap between different ontologies
            If any values are also used as ontology keys
        """
        # Check that all keys are valid ontologies
        invalid_ontologies = set(v.keys()) - set(ONTOLOGIES_LIST)
        if invalid_ontologies:
            raise ValueError(
                f"Invalid ontologies: {', '.join(invalid_ontologies)}. "
                f"Must be one of: {', '.join(ONTOLOGIES_LIST)}"
            )

        # Check that values don't overlap between keys and aren't used as keys
        all_values = set()
        keys = set(v.keys())
        for key, values in v.items():
            # Check for overlap with other values
            overlap = values & all_values
            if overlap:
                raise ValueError(
                    f"Found overlapping values {overlap} under multiple ontologies"
                )
            # Check for overlap with keys
            key_overlap = values & keys
            if key_overlap:
                raise ValueError(
                    f"Found values {key_overlap} that are also used as ontology keys"
                )
            all_values.update(values)

        return v


def _create_alias_mapping(ontology_dict: Dict[str, Set[str]]) -> Dict[str, str]:
    """Create a mapping from aliases to canonical ontology names.

    Only creates mappings for the aliases specified in the input dictionary.
    Does not include mappings for canonical names to themselves.

    Parameters
    ----------
    ontology_dict : Dict[str, Set[str]]
        Dictionary mapping ontologies to their aliases

    Returns
    -------
    Dict[str, str]
        Dictionary mapping each alias to its canonical ontology name
    """
    mapping = {}
    for ontology, aliases in ontology_dict.items():
        # Only map aliases to canonical names
        for alias in aliases:
            mapping[alias] = ontology
    return mapping


def _log_ontology_updates(
    alias_mapping: Dict[str, str], species_ontologies: Set[str]
) -> None:
    """Log which ontology aliases will be updated.

    Parameters
    ----------
    alias_mapping : Dict[str, str]
        Dictionary mapping old ontology names to new ones
    species_ontologies : Set[str]
        Set of ontology names present in the species identifiers

    Raises
    ------
    ValueError
        If there is no overlap between the aliases and species ontologies
    """
    # Find which aliases are present in the species data
    updatable_aliases = set(alias_mapping.keys()) & species_ontologies
    if not updatable_aliases:
        raise ValueError(
            "The set of ontologies in the species identifiers and aliases do not overlap. "
            "Please provide an updated aliases dict."
        )

    # Log which ontologies will be updated
    updates = [f"{old} -> {alias_mapping[old]}" for old in updatable_aliases]
    logger.info(f"Updating the following ontologies: {', '.join(updates)}")



"""Module to contain all constants used for representing and working with networks"""

from __future__ import annotations

from types import SimpleNamespace


from napistu.constants import SBML_DFS
from napistu.constants import SBOTERM_NAMES

NAPISTU_GRAPH = SimpleNamespace(VERTICES="vertices", EDGES="edges", METADATA="metadata")

GRAPH_DIRECTEDNESS = SimpleNamespace(DIRECTED="directed", UNDIRECTED="undirected")

GRAPH_RELATIONSHIPS = SimpleNamespace(
    ANCESTORS="ancestors",
    CHILDREN="children",
    DESCENDANTS="descendants",
    FOCAL="focal",
    PARENTS="parents",
)

NAPISTU_GRAPH_VERTICES = SimpleNamespace(
    NAME="name",  # internal name
    NODE_NAME="node_name",  # human readable name
    NODE_TYPE="node_type",  # type of node (species or reaction)
)

NAPISTU_GRAPH_EDGES = SimpleNamespace(
    DIRECTED="directed",
    DIRECTION="direction",
    FROM="from",
    R_ID=SBML_DFS.R_ID,
    R_ISREVERSIBLE=SBML_DFS.R_ISREVERSIBLE,
    SBO_TERM=SBML_DFS.SBO_TERM,
    SBO_NAME="sbo_name",
    SC_DEGREE="sc_degree",
    SC_PARENTS="sc_parents",
    SC_CHILDREN="sc_children",
    SPECIES_TYPE="species_type",
    STOICHIOMETRY=SBML_DFS.STOICHIOMETRY,
    TO="to",
    UPSTREAM_WEIGHT="upstream_weight",
    WEIGHT="weight",
)

NAPISTU_GRAPH_REQUIRED_EDGE_VARS = {
    NAPISTU_GRAPH_EDGES.FROM,
    NAPISTU_GRAPH_EDGES.TO,
    NAPISTU_GRAPH_EDGES.DIRECTION,
}

NAPISTU_GRAPH_NODE_TYPES = SimpleNamespace(SPECIES="species", REACTION="reaction")

VALID_NAPISTU_GRAPH_NODE_TYPES = [
    NAPISTU_GRAPH_NODE_TYPES.REACTION,
    NAPISTU_GRAPH_NODE_TYPES.SPECIES,
]

# translating an SBML_dfs -> NapistuGraph

GRAPH_WIRING_APPROACHES = SimpleNamespace(
    BIPARTITE="bipartite", REGULATORY="regulatory", SURROGATE="surrogate"
)

VALID_GRAPH_WIRING_APPROACHES = list(GRAPH_WIRING_APPROACHES.__dict__.values())

GRAPH_WIRING_HIERARCHIES = {
    # three tiers with reactions in the middle
    # in a bipartite networks molecules are connected to reactions but not other molecules
    GRAPH_WIRING_APPROACHES.BIPARTITE: [
        [
            SBOTERM_NAMES.CATALYST,
            SBOTERM_NAMES.INHIBITOR,
            SBOTERM_NAMES.INTERACTOR,
            SBOTERM_NAMES.MODIFIER,
            SBOTERM_NAMES.REACTANT,
            SBOTERM_NAMES.STIMULATOR,
        ],
        [NAPISTU_GRAPH_NODE_TYPES.REACTION],
        [SBOTERM_NAMES.MODIFIED, SBOTERM_NAMES.PRODUCT],
    ],
    # the regulatory graph defines a hierarchy of upstream and downstream
    # entities in a reaction
    # modifier/stimulator/inhibitor -> catalyst -> reactant -> reaction -> product
    GRAPH_WIRING_APPROACHES.REGULATORY: [
        [SBOTERM_NAMES.INHIBITOR, SBOTERM_NAMES.MODIFIER, SBOTERM_NAMES.STIMULATOR],
        [SBOTERM_NAMES.CATALYST],
        [SBOTERM_NAMES.INTERACTOR, SBOTERM_NAMES.REACTANT],
        [NAPISTU_GRAPH_NODE_TYPES.REACTION],
        [SBOTERM_NAMES.MODIFIED, SBOTERM_NAMES.PRODUCT],
    ],
    # an alternative layout to regulatory where enyzmes are downstream of substrates.
    # this doesn't make much sense from a regulatory perspective because
    # enzymes modify substrates not the other way around. but, its what one might
    # expect if catalysts are a surrogate for reactions as is the case for metabolic
    # network layouts
    GRAPH_WIRING_APPROACHES.SURROGATE: [
        [SBOTERM_NAMES.INHIBITOR, SBOTERM_NAMES.MODIFIER, SBOTERM_NAMES.STIMULATOR],
        [SBOTERM_NAMES.INTERACTOR, SBOTERM_NAMES.REACTANT],
        [SBOTERM_NAMES.CATALYST],
        [NAPISTU_GRAPH_NODE_TYPES.REACTION],
        [SBOTERM_NAMES.MODIFIED, SBOTERM_NAMES.PRODUCT],
    ],
}

# when should reaction vertices be excluded from the graph?

DROP_REACTIONS_WHEN = SimpleNamespace(
    ALWAYS="always",
    # if there are 2 participants
    EDGELIST="edgelist",
    # if there are 2 participants which are both "interactor"
    SAME_TIER="same_tier",
)

VALID_DROP_REACTIONS_WHEN = list(DROP_REACTIONS_WHEN.__dict__.values())

# adding weights to NapistuGraph

NAPISTU_WEIGHTING_STRATEGIES = SimpleNamespace(
    CALIBRATED="calibrated", MIXED="mixed", TOPOLOGY="topology", UNWEIGHTED="unweighted"
)

VALID_WEIGHTING_STRATEGIES = [
    NAPISTU_WEIGHTING_STRATEGIES.UNWEIGHTED,
    NAPISTU_WEIGHTING_STRATEGIES.TOPOLOGY,
    NAPISTU_WEIGHTING_STRATEGIES.MIXED,
    NAPISTU_WEIGHTING_STRATEGIES.CALIBRATED,
]

# edge reversal

NAPISTU_GRAPH_EDGE_DIRECTIONS = SimpleNamespace(
    FORWARD="forward", REVERSE="reverse", UNDIRECTED="undirected"
)

EDGE_REVERSAL_ATTRIBUTE_MAPPING = {
    NAPISTU_GRAPH_EDGES.FROM: NAPISTU_GRAPH_EDGES.TO,
    NAPISTU_GRAPH_EDGES.TO: NAPISTU_GRAPH_EDGES.FROM,
    NAPISTU_GRAPH_EDGES.SC_PARENTS: NAPISTU_GRAPH_EDGES.SC_CHILDREN,
    NAPISTU_GRAPH_EDGES.SC_CHILDREN: NAPISTU_GRAPH_EDGES.SC_PARENTS,
    NAPISTU_GRAPH_EDGES.WEIGHT: NAPISTU_GRAPH_EDGES.UPSTREAM_WEIGHT,
    NAPISTU_GRAPH_EDGES.UPSTREAM_WEIGHT: NAPISTU_GRAPH_EDGES.WEIGHT,
    # Note: stoichiometry requires special handling (* -1)
}

# Direction enum values
EDGE_DIRECTION_MAPPING = {
    NAPISTU_GRAPH_EDGE_DIRECTIONS.FORWARD: NAPISTU_GRAPH_EDGE_DIRECTIONS.REVERSE,
    NAPISTU_GRAPH_EDGE_DIRECTIONS.REVERSE: NAPISTU_GRAPH_EDGE_DIRECTIONS.FORWARD,
    NAPISTU_GRAPH_EDGE_DIRECTIONS.UNDIRECTED: NAPISTU_GRAPH_EDGE_DIRECTIONS.UNDIRECTED,  # unchanged
}

# Net edge direction
NET_POLARITY = SimpleNamespace(
    LINK_POLARITY="link_polarity",
    NET_POLARITY="net_polarity",
    ACTIVATION="activation",
    INHIBITION="inhibition",
    AMBIGUOUS="ambiguous",
    AMBIGUOUS_ACTIVATION="ambiguous activation",
    AMBIGUOUS_INHIBITION="ambiguous inhibition",
)

VALID_LINK_POLARITIES = [
    NET_POLARITY.ACTIVATION,
    NET_POLARITY.INHIBITION,
    NET_POLARITY.AMBIGUOUS,
]

VALID_NET_POLARITIES = [
    NET_POLARITY.ACTIVATION,
    NET_POLARITY.INHIBITION,
    NET_POLARITY.AMBIGUOUS,
    NET_POLARITY.AMBIGUOUS_ACTIVATION,
    NET_POLARITY.AMBIGUOUS_INHIBITION,
]

NEIGHBORHOOD_NETWORK_TYPES = SimpleNamespace(
    DOWNSTREAM="downstream", HOURGLASS="hourglass", UPSTREAM="upstream"
)

VALID_NEIGHBORHOOD_NETWORK_TYPES = [
    NEIGHBORHOOD_NETWORK_TYPES.DOWNSTREAM,
    NEIGHBORHOOD_NETWORK_TYPES.HOURGLASS,
    NEIGHBORHOOD_NETWORK_TYPES.UPSTREAM,
]

# weighting networks and transforming attributes

WEIGHTING_SPEC = SimpleNamespace(
    TABLE="table",
    VARIABLE="variable",
    TRANSFORMATION="trans",
)

DEFAULT_WT_TRANS = "identity"

DEFINED_WEIGHT_TRANSFORMATION = {
    DEFAULT_WT_TRANS: "_wt_transformation_identity",
    "string": "_wt_transformation_string",
    "string_inv": "_wt_transformation_string_inv",
}

SCORE_CALIBRATION_POINTS_DICT = {
    "weights": {"strong": 3, "good": 7, "okay": 20, "weak": 40},
    "string_wt": {"strong": 950, "good": 400, "okay": 230, "weak": 150},
}

SOURCE_VARS_DICT = {"string_wt": 10}

# network propagation
NET_PROPAGATION_DEFS = SimpleNamespace(PERSONALIZED_PAGERANK="personalized_pagerank")

# null distributions
NULL_STRATEGIES = SimpleNamespace(
    UNIFORM="uniform",
    PARAMETRIC="parametric",
    VERTEX_PERMUTATION="vertex_permutation",
    EDGE_PERMUTATION="edge_permutation",
)

VALID_NULL_STRATEGIES = NULL_STRATEGIES.__dict__.values()

PARAMETRIC_NULL_DEFAULT_DISTRIBUTION = "norm"

# masks

MASK_KEYWORDS = SimpleNamespace(
    ATTR="attr",
)

NEIGHBORHOOD_DICT_KEYS = SimpleNamespace(
    GRAPH="graph",
    VERTICES="vertices",
    EDGES="edges",
    REACTION_SOURCES="reaction_sources",
    NEIGHBORHOOD_PATH_ENTITIES="neighborhood_path_entities",
)

DISTANCES = SimpleNamespace(
    # core attributes of precomputed distances
    SC_ID_ORIGIN="sc_id_origin",
    SC_ID_DEST="sc_id_dest",
    PATH_LENGTH="path_length",
    PATH_UPSTREAM_WEIGHT="path_upstream_weight",
    PATH_WEIGHT="path_weight",
    # other attributes associated with paths/distances
    FINAL_FROM="final_from",
    FINAL_TO="final_to",
)



# GCS constants
from __future__ import annotations

from types import SimpleNamespace

from napistu.constants import NAPISTU_STANDARD_OUTPUTS

GCS_SUBASSET_NAMES = SimpleNamespace(
    SBML_DFS="sbml_dfs",
    NAPISTU_GRAPH="napistu_graph",
    SPECIES_IDENTIFIERS="species_identifiers",
    REACTIONS_SOURCE_TOTAL_COUNTS="reactions_source_total_counts",
    PRECOMPUTED_DISTANCES="precomputed_distances",
)


GCS_FILETYPES = SimpleNamespace(
    SBML_DFS="sbml_dfs.pkl",
    NAPISTU_GRAPH="napistu_graph.pkl",
    SPECIES_IDENTIFIERS=NAPISTU_STANDARD_OUTPUTS.SPECIES_IDENTIFIERS,
    REACTIONS_SOURCE_TOTAL_COUNTS=NAPISTU_STANDARD_OUTPUTS.REACTIONS_SOURCE_TOTAL_COUNTS,
    PRECOMPUTED_DISTANCES="precomputed_distances.parquet",
)


GCS_ASSETS = SimpleNamespace(
    PROJECT="calico-public-data",
    BUCKET="calico-cpr-public",
    ASSETS={
        "test_pathway": {
            "file": "test_pathway.tar.gz",
            "subassets": {
                GCS_SUBASSET_NAMES.SBML_DFS: GCS_FILETYPES.SBML_DFS,
                GCS_SUBASSET_NAMES.NAPISTU_GRAPH: GCS_FILETYPES.NAPISTU_GRAPH,
                GCS_SUBASSET_NAMES.SPECIES_IDENTIFIERS: GCS_FILETYPES.SPECIES_IDENTIFIERS,
                GCS_SUBASSET_NAMES.PRECOMPUTED_DISTANCES: GCS_FILETYPES.PRECOMPUTED_DISTANCES,
                GCS_SUBASSET_NAMES.REACTIONS_SOURCE_TOTAL_COUNTS: GCS_FILETYPES.REACTIONS_SOURCE_TOTAL_COUNTS,
            },
            "public_url": "https://storage.googleapis.com/shackett-napistu-public/test_pathway.tar.gz",
        },
        "human_consensus": {
            "file": "human_consensus.tar.gz",
            "subassets": {
                GCS_SUBASSET_NAMES.SBML_DFS: GCS_FILETYPES.SBML_DFS,
                GCS_SUBASSET_NAMES.NAPISTU_GRAPH: GCS_FILETYPES.NAPISTU_GRAPH,
                GCS_SUBASSET_NAMES.SPECIES_IDENTIFIERS: GCS_FILETYPES.SPECIES_IDENTIFIERS,
                GCS_SUBASSET_NAMES.REACTIONS_SOURCE_TOTAL_COUNTS: GCS_FILETYPES.REACTIONS_SOURCE_TOTAL_COUNTS,
            },
            "public_url": "https://storage.googleapis.com/shackett-napistu-public/human_consensus.tar.gz",
        },
        "human_consensus_w_distances": {
            "file": "human_consensus_w_distances.tar.gz",
            "subassets": {
                GCS_SUBASSET_NAMES.SBML_DFS: GCS_FILETYPES.SBML_DFS,
                GCS_SUBASSET_NAMES.NAPISTU_GRAPH: GCS_FILETYPES.NAPISTU_GRAPH,
                GCS_SUBASSET_NAMES.SPECIES_IDENTIFIERS: GCS_FILETYPES.SPECIES_IDENTIFIERS,
                GCS_SUBASSET_NAMES.PRECOMPUTED_DISTANCES: GCS_FILETYPES.PRECOMPUTED_DISTANCES,
                GCS_SUBASSET_NAMES.REACTIONS_SOURCE_TOTAL_COUNTS: GCS_FILETYPES.REACTIONS_SOURCE_TOTAL_COUNTS,
            },
            "public_url": "https://storage.googleapis.com/shackett-napistu-public/human_consensus_w_distances.tar.gz",
        },
        "reactome_members": {
            "file": "external_pathways/external_pathways_reactome_neo4j_members.csv",
            "subassets": None,
            "public_url": "https://storage.googleapis.com/calico-cpr-public/external_pathways/external_pathways_reactome_neo4j_members.csv",
        },
        "reactome_xrefs": {
            "file": "external_pathways/external_pathways_reactome_neo4j_crossref.csv",
            "subassets": None,
            "public_url": "https://storage.googleapis.com/calico-cpr-public/external_pathways/external_pathways_reactome_neo4j_crossref.csv",
        },
    },
)


INIT_DATA_DIR_MSG = "The `data_dir` {data_dir} does not exist."



"""
MCP (Model Context Protocol) Server for Napistu.

This module requires optional dependencies. Install with:
pip install napistu[mcp]
"""

import asyncio
from typing import Dict, Any

__all__ = ["start_server", "register_object", "is_available"]

# Check if MCP dependencies are available
try:
    __import__("mcp")
    is_available = True
except ImportError:
    is_available = False

if is_available:
    from napistu.mcp.server import create_server
    from napistu.mcp.profiles import get_profile
    from napistu.mcp.constants import MCP_PROFILES

    def start_server(
        profile_name: str = MCP_PROFILES.EXECUTION, **kwargs
    ) -> Dict[str, Any]:
        """
        Start an MCP server with a specific profile.

        Args:
            profile_name: Name of the profile ('execution', 'docs', or 'full')
            **kwargs: Additional configuration options

        Returns:
            Server control dictionary
        """
        profile = get_profile(profile_name, **kwargs)
        server = create_server(profile)

        # Start the server
        asyncio.create_task(server.start())

        # Return control interface
        return {
            "status": "running",
            "server": server,
            "profile": profile_name,
            "stop": server.stop,
            "register_object": (
                register_object if profile.get_config()["enable_execution"] else None
            ),
        }

    # Helper function for registering objects with a running server
    def register_object(name, obj):
        """Register an object with the execution component."""
        from .execution import register_object as _register

        return _register(name, obj)

else:
    # Stubs for when MCP is not available
    def start_server(*args, **kwargs):
        raise ImportError(
            "MCP support not installed. Install with 'pip install napistu[mcp]'"
        )

    def register_object(*args, **kwargs):
        raise ImportError(
            "MCP support not installed. Install with 'pip install napistu[mcp]'"
        )



import pandas as pd

from napistu.network import paths
from napistu.network import ng_utils
from napistu.constants import SBML_DFS
from napistu.network.constants import (
    NAPISTU_GRAPH_EDGES,
    NET_POLARITY,
)


def test_shortest_paths(sbml_dfs, napistu_graph, napistu_graph_undirected):
    species = sbml_dfs.species
    source_species = species[species[SBML_DFS.S_NAME] == "NADH"]
    dest_species = species[species[SBML_DFS.S_NAME] == "NAD+"]
    target_species_paths = ng_utils.compartmentalize_species_pairs(
        sbml_dfs, source_species.index.tolist(), dest_species.index.tolist()
    )

    (
        all_shortest_reaction_paths_df,
        _,
        _,
        _,
    ) = paths.find_all_shortest_reaction_paths(
        napistu_graph,
        sbml_dfs,
        target_species_paths,
        weight_var=NAPISTU_GRAPH_EDGES.WEIGHT,
    )

    # undirected graph
    (
        all_shortest_reaction_paths_df,
        _,
        _,
        _,
    ) = paths.find_all_shortest_reaction_paths(
        napistu_graph_undirected,
        sbml_dfs,
        target_species_paths,
        weight_var=NAPISTU_GRAPH_EDGES.WEIGHT,
    )

    assert all_shortest_reaction_paths_df.shape[0] == 3


def test_net_polarity():
    polarity_series = pd.Series(
        [NET_POLARITY.AMBIGUOUS, NET_POLARITY.AMBIGUOUS],
        index=[0, 1],
        name=NET_POLARITY.LINK_POLARITY,
    )
    assert all(
        [
            x == NET_POLARITY.AMBIGUOUS
            for x in paths._calculate_net_polarity(polarity_series)
        ]
    )

    polarity_series = pd.Series(
        [
            NET_POLARITY.ACTIVATION,
            NET_POLARITY.INHIBITION,
            NET_POLARITY.INHIBITION,
            NET_POLARITY.AMBIGUOUS,
        ],
        index=range(0, 4),
        name=NET_POLARITY.LINK_POLARITY,
    )
    assert paths._calculate_net_polarity(polarity_series) == [
        NET_POLARITY.ACTIVATION,
        NET_POLARITY.INHIBITION,
        NET_POLARITY.ACTIVATION,
        NET_POLARITY.AMBIGUOUS_ACTIVATION,
    ]
    assert (
        paths._terminal_net_polarity(polarity_series)
        == NET_POLARITY.AMBIGUOUS_ACTIVATION
    )



import logging
import os
import pathlib

from napistu.gcs.constants import INIT_DATA_DIR_MSG

logger = logging.getLogger(__name__)


def _initialize_data_dir(data_dir: str, init_msg: str = INIT_DATA_DIR_MSG) -> None:
    """Create a data directory if it doesn't exist."""

    if not os.path.isdir(data_dir):

        logger.warning(init_msg.format(data_dir=data_dir))

        # Artifact directory not found; creating {parentdir}")
        logger.warning(f"Trying to create {data_dir}")
        pathlib.Path(data_dir).mkdir(parents=True, exist_ok=True)

    return None



from __future__ import annotations

import logging

import pandas as pd
from napistu.rpy2 import (
    require_rpy2,
    report_r_exceptions,
)

from napistu.rpy2.callr import bioconductor_org_r_function, r_dataframe_to_pandas

from napistu.constants import ONTOLOGIES
from napistu.rpy2.constants import BIOC_VALID_EXPANDED_SPECIES_ONTOLOGIES
from napistu.rpy2.constants import BIOC_ONTOLOGY_MAPPING
from napistu.rpy2.constants import BIOC_NOMENCLATURE

logger = logging.getLogger(__name__)


@require_rpy2
@report_r_exceptions
def create_bioconductor_mapping_tables(
    mappings: set[str], species: str, r_paths: str | None = None
) -> dict[str, pd.DataFrame]:
    """Create Bioconductor Mapping Tables.

    Creating a dictionary of mappings between entrez and other ontologies.

    Parameters
    ----------
    mappings : set[str]
        A set of ontologies to work with. The valid ontologies are:
        "ensembl_gene", "ensembl_transcript", and "uniprot".
    species : str
        The organismal species that we are working with (e.g., Homo sapiens).
    r_paths : str | None, optional
        Optional path to a library of R packages.

    Returns
    -------
    dict[str, pd.DataFrame]
        A table of entrez ids, and tables mapping from each ontology in "mappings" to entrez.

    Raises
    ------
    ValueError
        If any of the requested mappings are not supported
    """
    logger.info(
        "Creating mapping tables from entrez genes to/from %s", ", ".join(mappings)
    )

    invalid_mappings = set(mappings).difference(BIOC_VALID_EXPANDED_SPECIES_ONTOLOGIES)
    if len(invalid_mappings) > 0:
        raise ValueError(
            f"{len(invalid_mappings)} mappings could not be created: {', '.join(invalid_mappings)}.\n"
            f"The valid mappings are {', '.join(BIOC_VALID_EXPANDED_SPECIES_ONTOLOGIES)}"
        )

    mappings_dict = {}

    # Create mapping tables for each requested ontology
    for ontology in mappings:
        mappings_dict[ontology] = _create_single_mapping(ontology, species, r_paths)

    return mappings_dict


def _create_single_mapping(
    ontology: str, species: str, r_paths: str | None = None
) -> pd.DataFrame:
    """Create a single mapping table for a given ontology.

    Parameters
    ----------
    ontology : str
        The ontology to map (e.g. ENSEMBL_GENE, UNIPROT)
    species : str
        The organismal species to map
    r_paths : str | None, optional
        Optional path to R packages directory

    Returns
    -------
    pd.DataFrame
        DataFrame containing the mapping between entrez and the target ontology
    """

    if ontology not in BIOC_ONTOLOGY_MAPPING:
        raise ValueError(f"Unsupported ontology: {ontology}")

    table_name, column_name = BIOC_ONTOLOGY_MAPPING[ontology]

    df = r_dataframe_to_pandas(
        bioconductor_org_r_function(table_name, species, r_paths=r_paths)
    )

    # Drop chromosome column if this is the chromosome table
    # this was only introduced so we had a table with 1 row per unique entrez id
    if table_name == BIOC_NOMENCLATURE.CHR_TBL:
        df = df.drop(BIOC_NOMENCLATURE.CHROMOSOME, axis=1)

    # Rename columns and set index
    df = df.rename(
        columns={
            BIOC_NOMENCLATURE.NCBI_ENTREZ_GENE: ONTOLOGIES.NCBI_ENTREZ_GENE,
            column_name: ontology,
        }
    ).set_index(ONTOLOGIES.NCBI_ENTREZ_GENE)

    return df



import os
import logging

import pandas as pd

from napistu import utils
from napistu.identifiers import Identifiers
from napistu.constants import IDENTIFIERS, SBML_DFS
from napistu.ingestion.constants import (
    PSI_MI_DEFS,
    PSI_MI_INTACT_FTP_URL,
    PSI_MI_INTACT_SPECIES_TO_BASENAME,
)

logger = logging.getLogger(__name__)


def download_intact_xmls(
    output_dir_path: str,
    latin_species: str,
    overwrite: bool = False,
) -> None:
    """
    Download IntAct Species

    Download the PSM-30 XML files from IntAct for a species of interest.

    Parameters
    ----------
    output_dir_path (str):
        Local directory to create an unzip files into
    latin_species (str):
        The species name (Genus species) to work with
    overwrite (bool):
        Overwrite an existing output directory. Default: False
    """

    intact_species_basename = _get_intact_species_basename(latin_species)

    intact_species_url = os.path.join(
        PSI_MI_INTACT_FTP_URL, f"{intact_species_basename}.zip"
    )

    logger.info(f"Downloading and unzipping {intact_species_url}")

    utils.download_and_extract(
        intact_species_url,
        output_dir_path=output_dir_path,
        download_method="ftp",
        overwrite=overwrite,
    )


def create_species_df(
    raw_species_df: pd.DataFrame,
    raw_species_identifiers_df: pd.DataFrame,
    latin_species: str,
):
    """
    Create a species dataframe from the raw species dataframe and the raw species identifiers dataframe.

    Parameters
    ----------
    raw_species_df : pd.DataFrame
        The raw species dataframe.
    raw_species_identifiers_df : pd.DataFrame
        The raw species identifiers dataframe.
    latin_species : str
        The latin species name in the format "Genus species".

    Returns
    -------
    lookup_table : pd.Series
        A lookup table mapping study_id and interactor_id to the molecular species name.
    species_df : pd.DataFrame
        The molecular species dataframe.
    """

    # filter to just matchees within the same species
    intact_species_basename = _get_intact_species_basename(latin_species)

    # all partipants
    valid_species_mask = raw_species_df[PSI_MI_DEFS.INTERACTOR_LABEL].str.endswith(
        intact_species_basename
    )

    logger.info(
        f"Retaining {sum(valid_species_mask)} interactors which are from {intact_species_basename} and removing {sum(~valid_species_mask)} interactors which are not."
    )
    valid_species_df = raw_species_df.loc[valid_species_mask]

    species_df = (
        raw_species_identifiers_df.merge(
            valid_species_df[
                [
                    PSI_MI_DEFS.INTERACTOR_ID,
                    PSI_MI_DEFS.STUDY_ID,
                    PSI_MI_DEFS.INTERACTOR_LABEL,
                ]
            ],
            on=[PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTOR_ID],
            how="inner",
        )[
            [
                PSI_MI_DEFS.INTERACTOR_LABEL,
                IDENTIFIERS.ONTOLOGY,
                IDENTIFIERS.IDENTIFIER,
                IDENTIFIERS.BQB,
            ]
        ]
        .drop_duplicates()
        .groupby([PSI_MI_DEFS.INTERACTOR_LABEL])
        .apply(lambda x: Identifiers(x.to_dict(orient="records")), include_groups=False)
        .rename(SBML_DFS.S_IDENTIFIERS)
        .rename_axis(SBML_DFS.S_NAME)
        .reset_index()
    )

    lookup_table = valid_species_df.set_index(
        [PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTOR_ID]
    )[PSI_MI_DEFS.INTERACTOR_LABEL]

    return lookup_table, species_df


def _get_intact_species_basename(latin_species: str) -> str:

    if latin_species not in PSI_MI_INTACT_SPECIES_TO_BASENAME.keys():
        raise ValueError(
            f"The provided species {latin_species} did not match any of the species in INTACT_SPECIES_TO_BASENAME: "
            f"{', '.join(PSI_MI_INTACT_SPECIES_TO_BASENAME.keys())}"
            "If this is a species supported by IntAct please add the species to the PSI_MI_INTACT_SPECIES_TO_BASENAME dictionary."
        )

    return PSI_MI_INTACT_SPECIES_TO_BASENAME[latin_species]



from __future__ import annotations

import os
import re

import pandas as pd
import requests
from napistu.ingestion.constants import IDENTIFIERS_ETL_SBO_URL
from napistu.ingestion.constants import IDENTIFIERS_ETL_YEAST_FIELDS
from napistu.ingestion.constants import IDENTIFIERS_ETL_YEAST_URL
from napistu.ingestion.constants import IDENTIFIERS_ETL_YEAST_HEADER_REGEX


def read_yeast_identifiers(url: str = IDENTIFIERS_ETL_YEAST_URL):
    """Read Yeast Identifiers
    Generate a pd.DataFrame which maps between yeast identifiers including
    common and systematic (OLN) names, as well as Swiss-Prot and SGD identifiers.

    Params:
        url (str): url to the identifier file
    Returns:
        pd.DataFrame with one row per gene
    """
    response = requests.get(url).text

    yeast_id_list = list()
    break_line_hit = 0
    for line in response.splitlines():
        if re.match(IDENTIFIERS_ETL_YEAST_HEADER_REGEX, line):
            # find start and end of header indicated by a line of underscores
            break_line_hit += 1
            continue

        if break_line_hit >= 2:
            if line == "":
                # reached the end
                break

            # split each line into a list of fields, the only optional field is 3d
            # all white spaces are space
            line = re.sub(" +", " ", line)
            line = re.sub("; ", ";", line)
            # remove pol and gag designations from transposons since they are an unnecessary extra field
            line = re.sub("(-[0-9]) (GAG)|(POL)", "\\1", line)

            line = line.split()

            if line[6] != "(3)":
                # if no 3D field is present then create one
                line.insert(6, "none")

            # split common fields into a separate list
            common_list = line[0].split(";")
            line[0] = common_list[0]
            line.insert(1, common_list)

            if len(line) != 9:
                raise ValueError(
                    "the yeast id file could not be read; all entries should have 8 fields"
                )

            yeast_id_list.append(dict(zip(IDENTIFIERS_ETL_YEAST_FIELDS, line)))

    return pd.DataFrame(yeast_id_list)


def read_sbo_ontology(
    url: str = IDENTIFIERS_ETL_SBO_URL, verbose: bool = False
) -> pd.DataFrame:
    """Read SBO Ontology
    Read the Systems Biology Ontology (SBO) identifiers and reformat the obo results into a pd.DataFrame.

    Params:
        url (str): url to the obo specification file
        verbose (bool): throw warnings when attributes are overwritten
    Returns:
        pd.DataFrame
    """

    # save the obo file locally
    tmp_file = os.path.join("/tmp", "sbo.obo")
    r = requests.get(url, allow_redirects=True)
    open(tmp_file, "wb").write(r.content)

    with open(tmp_file) as sbo:
        sbo_dict = dict()
        current_id = None
        in_header = True
        for line in sbo:
            # skip the header
            if line == "[Term]\n":
                in_header = False
                continue
            if in_header:
                continue

            line_entries = line.split(":", 1)

            if len(line_entries) == 2:
                entry_type = line_entries[0]
                entry_value = line_entries[1].strip()

                # drop type defs
                if (
                    (current_id is not None)
                    and (entry_type != "id")
                    and (re.match("SBO", current_id) is None)
                ):
                    continue

                # clean-up definitions
                if entry_type == "is_a":
                    entry_value = re.match("SBO:[0-9]+", entry_value)[0]

                # if a new id has been reached then initilize a new dict and
                # update current id

                if entry_type == "id":
                    current_id = entry_value
                    if re.match("SBO", current_id) is not None:
                        sbo_dict[current_id] = {"is_a": []}
                    continue

                if entry_type == "is_a":
                    sbo_dict[current_id]["is_a"].append(entry_value)
                else:
                    # add a new entry
                    if (entry_type in sbo_dict[current_id].keys()) and verbose:
                        print(
                            f"2+ {entry_type} entries were found for {current_id}, only one value should be present "
                        )
                    sbo_dict[current_id][entry_type] = entry_value

    sbo_df = pd.DataFrame(sbo_dict).T

    obsolete_terms = set(
        sbo_df["name"][sbo_df["name"].str.match("obsolete")].index.tolist()
    )
    sbo_df["is_obsolete"] = [
        (x in obsolete_terms) | (len(set(y).intersection(obsolete_terms)) > 0)
        for x, y in zip(sbo_df.index, sbo_df["is_a"])
    ]

    sbo_df = sbo_df[["name", "comment", "is_a", "is_obsolete"]]
    sbo_df.index.name = "sbo_term"
    sbo_df = sbo_df.reset_index()

    return sbo_df



"""
MCP (Model Context Protocol) Server CLI for Napistu.
"""

import asyncio
import click
import click_logging
import logging

import napistu
from napistu.mcp.server import start_mcp_server
from napistu.mcp.client import (
    check_server_health,
    print_health_status,
    list_server_resources,
    read_server_resource,
)
from napistu.mcp.config import (
    validate_server_config_flags,
    validate_client_config_flags,
    server_config_options,
    client_config_options,
    local_server_config,
    local_client_config,
    production_client_config,
)

logger = logging.getLogger(napistu.__name__)
click_logging.basic_config(logger)


@click.group()
def cli():
    """The Napistu MCP (Model Context Protocol) Server CLI"""
    pass


@click.group()
def server():
    """Start and manage MCP servers."""
    pass


@server.command(name="start")
@click.option(
    "--profile", type=click.Choice(["execution", "docs", "full"]), default="docs"
)
@server_config_options
@click_logging.simple_verbosity_option(logger)
def start_server(profile, production, local, host, port, server_name):
    """Start an MCP server with the specified profile."""
    try:
        config = validate_server_config_flags(
            local, production, host, port, server_name
        )

        click.echo("Starting server with configuration:")
        click.echo(f"  Profile: {profile}")
        click.echo(f"  Host: {config.host}")
        click.echo(f"  Port: {config.port}")
        click.echo(f"  Server Name: {config.server_name}")

        start_mcp_server(profile, config)

    except click.BadParameter as e:
        raise click.ClickException(str(e))


@server.command(name="local")
@click_logging.simple_verbosity_option(logger)
def start_local():
    """Start a local MCP server optimized for function execution."""
    config = local_server_config()
    click.echo("Starting local development server (execution profile)")
    click.echo(f"  Host: {config.host}")
    click.echo(f"  Port: {config.port}")
    click.echo(f"  Server Name: {config.server_name}")

    start_mcp_server("execution", config)


@server.command(name="full")
@click_logging.simple_verbosity_option(logger)
def start_full():
    """Start a full MCP server with all components enabled (local debugging)."""
    config = local_server_config()
    # Override server name for full profile
    config.server_name = "napistu-full"

    click.echo("Starting full development server (all components)")
    click.echo(f"  Host: {config.host}")
    click.echo(f"  Port: {config.port}")
    click.echo(f"  Server Name: {config.server_name}")

    start_mcp_server("full", config)


@cli.command()
@client_config_options
@click_logging.simple_verbosity_option(logger)
def health(production, local, host, port, https):
    """Quick health check of MCP server."""

    async def run_health_check():
        try:
            config = validate_client_config_flags(local, production, host, port, https)

            print("🏥 Napistu MCP Server Health Check")
            print("=" * 40)
            print(f"Server URL: {config.base_url}")
            print()

            health = await check_server_health(config)
            print_health_status(health)

        except click.BadParameter as e:
            raise click.ClickException(str(e))

    asyncio.run(run_health_check())


@cli.command()
@client_config_options
@click_logging.simple_verbosity_option(logger)
def resources(production, local, host, port, https):
    """List all available resources on the MCP server."""

    async def run_list_resources():
        try:
            config = validate_client_config_flags(local, production, host, port, https)

            print("📋 Napistu MCP Server Resources")
            print("=" * 40)
            print(f"Server URL: {config.base_url}")
            print()

            resources = await list_server_resources(config)

            if resources:
                print(f"Found {len(resources)} resources:")
                for resource in resources:
                    print(f"  📄 {resource.uri}")
                    if resource.name != resource.uri:
                        print(f"      Name: {resource.name}")
                    if hasattr(resource, "description") and resource.description:
                        print(f"      Description: {resource.description}")
                    print()
            else:
                print("❌ Could not retrieve resources")

        except click.BadParameter as e:
            raise click.ClickException(str(e))

    asyncio.run(run_list_resources())


@cli.command()
@click.argument("resource_uri")
@client_config_options
@click.option(
    "--output", type=click.File("w"), default="-", help="Output file (default: stdout)"
)
@click_logging.simple_verbosity_option(logger)
def read(resource_uri, production, local, host, port, https, output):
    """Read a specific resource from the MCP server."""

    async def run_read_resource():
        try:
            config = validate_client_config_flags(local, production, host, port, https)

            print(
                f"📖 Reading Resource: {resource_uri}",
                file=output if output.name != "<stdout>" else None,
            )
            print(
                f"Server URL: {config.base_url}",
                file=output if output.name != "<stdout>" else None,
            )
            print("=" * 50, file=output if output.name != "<stdout>" else None)

            content = await read_server_resource(resource_uri, config)

            if content:
                print(content, file=output)
            else:
                print(
                    "❌ Could not read resource",
                    file=output if output.name != "<stdout>" else None,
                )

        except click.BadParameter as e:
            raise click.ClickException(str(e))

    asyncio.run(run_read_resource())


@cli.command()
@click_logging.simple_verbosity_option(logger)
def compare():
    """Compare health between local development and production servers."""

    async def run_comparison():

        local_config = local_client_config()
        production_config = production_client_config()

        print("🔍 Local vs Production Server Comparison")
        print("=" * 50)

        print(f"\n📍 Local Server: {local_config.base_url}")
        local_health = await check_server_health(local_config)
        print_health_status(local_health)

        print(f"\n🌐 Production Server: {production_config.base_url}")
        production_health = await check_server_health(production_config)
        print_health_status(production_health)

        # Compare results
        print("\n📊 Comparison Summary:")
        if local_health and production_health:
            local_components = local_health.get("components", {})
            production_components = production_health.get("components", {})

            all_components = set(local_components.keys()) | set(
                production_components.keys()
            )

            for component in sorted(all_components):
                local_status = local_components.get(component, {}).get(
                    "status", "missing"
                )
                production_status = production_components.get(component, {}).get(
                    "status", "missing"
                )

                if local_status == production_status == "healthy":
                    icon = "✅"
                elif local_status != production_status:
                    icon = "⚠️ "
                else:
                    icon = "❌"

                print(
                    f"  {icon} {component}: Local={local_status}, Production={production_status}"
                )
        else:
            print("  ❌ Cannot compare - one or both servers unreachable")

    asyncio.run(run_comparison())


# Add commands to the CLI
cli.add_command(server)
cli.add_command(health)
cli.add_command(resources)
cli.add_command(read)
cli.add_command(compare)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import datetime
import logging
import os
import random
from io import StringIO
from typing import Iterable

import pandas as pd
import requests
from napistu import indices
from napistu import sbml_dfs_core
from napistu import utils
from napistu.consensus import construct_consensus_model
from napistu.consensus import construct_sbml_dfs_dict
from napistu.ingestion.constants import REACTOME_PATHWAY_INDEX_COLUMNS
from napistu.ingestion.constants import REACTOME_PATHWAY_LIST_COLUMNS
from napistu.ingestion.constants import REACTOME_PATHWAYS_URL
from napistu.ingestion.constants import REACTOME_SMBL_URL
from fs import open_fs

logger = logging.getLogger(__name__)


def reactome_sbml_download(output_dir_path: str, overwrite: bool = False):
    """
    Reactome SBML Download

    Download Reactome SBML (systems biology markup language) for all reactome species.

    Args:
        output_dir_path (str): Paths to a directory where .sbml files should be saved.
        overwrite (bool): Overwrite an existing output directory. Default: False
    """
    utils.download_and_extract(
        REACTOME_SMBL_URL,
        output_dir_path=output_dir_path,
        overwrite=overwrite,
    )
    # create the pathway index
    pw_index = _build_reactome_pw_index(output_dir_path, file_ext="sbml")

    # save as tsv
    out_fs = open_fs(output_dir_path)
    with out_fs.open("pw_index.tsv", "wb") as index_path:
        pw_index.to_csv(index_path, sep="\t", index=False)


# Functions useful to integrate reactome pathways into a consensus
def construct_reactome_consensus(
    pw_index_inp: str | indices.PWIndex,
    species: str | Iterable[str] | None = None,
    outdir: str | None = None,
    strict: bool = True,
) -> sbml_dfs_core.SBML_dfs:
    """Constructs a basic consensus model by merging all models from a pw_index

    Args:
        pw_index_inp (str | indices.PWIndex): PWIndex or uri pointing to PWIndex
        species (str | Iterable[str] | None): one or more species to filter by. Default: no filtering
        outdir (str | None, optional): output directory used to cache results. Defaults to None.
        strict (bool): should failure of loading any given model throw an exception? If False a warning is thrown.

    Returns:
        sbml_dfs_core.SBML_dfs: A consensus SBML
    """
    if isinstance(pw_index_inp, str):
        pw_index = indices.adapt_pw_index(pw_index_inp, species=species, outdir=outdir)
    elif isinstance(pw_index_inp, indices.PWIndex):
        pw_index = pw_index_inp
    else:
        raise ValueError("pw_index_inp needs to be a PWIndex or a str to a location.")
    if outdir is not None:
        construct_sbml_dfs_dict_fkt = utils.pickle_cache(
            os.path.join(outdir, "model_pool.pkl")
        )(construct_sbml_dfs_dict)
        construct_consensus_model_fkt = utils.pickle_cache(
            os.path.join(outdir, "consensus.pkl")
        )(construct_consensus_model)
    else:
        construct_sbml_dfs_dict_fkt = construct_sbml_dfs_dict
        construct_consensus_model_fkt = construct_consensus_model

    sbml_dfs_dict = construct_sbml_dfs_dict_fkt(pw_index, strict)
    consensus_model = construct_consensus_model_fkt(sbml_dfs_dict, pw_index)
    return consensus_model


def _build_reactome_pw_index(
    output_dir: str,
    file_ext: str,
    species_filter: Iterable[str] | None = None,
) -> pd.DataFrame:
    """Build a reactome pathway index

    Builds the index based on available files and cross-checkes it with the
    expected reactome pathway list.

    Args:
        output_dir (str): File directory
        file_ext (str): File extension
        species_filter (Optional[Iterable[str]], optional): Filter the expected
            pathway list based on a list of species. Eg in cases only one species available. Defaults to None.

    Returns:
        pd.DataFrame: pathway index
    """
    # create the pathway index
    out_fs = open_fs(output_dir)
    all_files = [os.path.basename(f.path) for f in out_fs.glob(f"**/*.{file_ext}")]

    if len(all_files) == 0:
        raise ValueError(f"Zero files in {output_dir} have the {file_ext} extension")

    pw_index = pd.DataFrame({"file": all_files}).assign(source="Reactome")
    pw_index["pathway_id"] = [os.path.splitext(x)[0] for x in pw_index["file"]]

    # test before merging
    pathway_list = _get_reactome_pathway_list()
    if species_filter is not None:
        pathway_list = pathway_list.loc[pathway_list["species"].isin(species_filter)]

    _check_reactome_pw_index(pw_index, pathway_list)
    pw_index = pw_index.merge(pathway_list)
    pw_index = pw_index[REACTOME_PATHWAY_INDEX_COLUMNS]
    pw_index["date"] = datetime.date.today().strftime("%Y%m%d")

    return pw_index


def _check_reactome_pw_index(pw_index: indices.PWIndex, reactome_pathway_list: list):
    """Compare local files defined in the pathway index to a list of Reactome's pathways."""

    # check extension in pw_index
    extn = set([os.path.splitext(x)[1] for x in pw_index["file"]])
    if len(extn) != 1:
        raise ValueError(
            f"Expected all files to have the same extension, but found extensions: {extn}"
        )
    if len(extn.intersection({".sbml"})) != 1:
        raise ValueError(
            f"Expected all files to have the .sbml extension, but found: {extn}"
        )
    extn_string = extn.pop()

    local_reactome_pws = set(pw_index["pathway_id"])
    remote_reactome_pws = set(reactome_pathway_list["pathway_id"])

    extra_local = local_reactome_pws.difference(remote_reactome_pws)
    if len(extra_local) != 0:
        n_samples = min(5, len(extra_local))
        local_str = ", ".join(random.sample(list(extra_local), n_samples))

        logger.warning(
            f"{len(extra_local)} Reactome {extn_string} files were detected "
            "which are not found in reactome.get_reactome_pathway_list(). "
            f"The include {local_str}. "
            "These files will be excluded from the pathway index"
        )

    extra_remote = remote_reactome_pws.difference(local_reactome_pws)

    if len(extra_remote) != 0:
        n_samples = min(5, len(extra_remote))
        remote_str = ", ".join(random.sample(list(extra_remote), n_samples))

        logger.warning(
            f"{len(extra_remote)} Reactome {extn_string} files were missing "
            "which should be present based on reactome.get_reactome_pathway_list(). "
            f"These include {remote_str}."
        )
    return None


def _get_reactome_pathway_list():
    """Reactome Pathway List
    Produce a pd.DataFrame listing all pathways in reactome and their internal ids

    Parameters:
        None

    Returns:
        pd.DataFrame containing pathway_id, name and species
    """
    page = requests.get(REACTOME_PATHWAYS_URL)
    if page.status_code != 200:
        raise ValueError(
            f"Reactome data could not be accessed at {REACTOME_PATHWAYS_URL}"
        )
    StringData = StringIO(page.content.decode())
    df = pd.read_csv(StringData, sep="\t", names=REACTOME_PATHWAY_LIST_COLUMNS)

    return df
